# The task of the classes and functions here is to convert
# JSON encoded instrument setups into an
from __future__ import print_function, unicode_literals, absolute_import, division
from collections import OrderedDict
from itertools import islice

from ..gtc.headers import create_header_from_telpars


def get_obsmode(setup_data):
    mode = setup_data["appdata"]["app"]
    if mode == "FullFrame":
        return FullFrame(setup_data)
    elif mode == "Windows":
        return Windows(setup_data)
    elif mode == "Drift":
        return Drift(setup_data)
    elif mode == "Idle":
        return Idle()
    else:
        raise ValueError("Unrecognised mode: {}".format(mode))


class ObsMode(object):
    def __init__(self, setup_data):
        """
        The base class for all HiPERCAM modes.

        Parameters
        ----------
        setup_data : dict
            Dictionary of HiPerCAM setup data
        """
        app_data = setup_data["appdata"]
        nu, ng, nr, ni, nz = app_data["multipliers"]
        dummy = app_data.get(
            "dummy_out", 0
        )  # works even if dummy not set in app, default 0
        fastclk = app_data.get("fast_clks", 0)
        oscany = app_data.get("oscany", 0)

        # look in dictionary to see if we should use GPS hardware
        # default is to assume GPS attached
        self.use_gps_hardware = setup_data.get("gps_attached", 1)

        clockfile = "hipercam.bclk"
        if not dummy:
            clockfile = "hipercam_se.bclk"
        elif fastclk:
            clockfile = "hipercam_fastclk.bclk"

        nodpattern = app_data.get("nodpattern", {})
        self.finite = app_data["numexp"]
        self.detpars = {
            "DET.RESTRICT": "F",
            "DET.SPEED": 0 if app_data["readout"] == "Slow" else 1,
            "DET.BINX1": app_data["xbin"],
            "DET.BINY1": app_data["ybin"],
            "DET.CLRCCD": "T" if app_data["clear"] else "F",
            "DET.NCLRS": 5,
            "DET.DUMMY": dummy,
            "DET.FASTCLK": fastclk,
            "DET.EXPLED": "T" if app_data["led_flsh"] else "F",
            "DET.GPS": "T",
            "DET.INCPRSCX": "T" if app_data["oscan"] else "F",
            "DET.INCOVSCY": "T" if oscany else "F",
            "DET.NSKIPS1": nu - 1,
            "DET.NSKIPS2": ng - 1,
            "DET.NSKIPS3": nr - 1,
            "DET.NSKIPS4": ni - 1,
            "DET.NSKIPS5": nz - 1,
            "DET.SEQ.CLKFILE": clockfile,
            "DET.SEQ1.DIT": app_data["exptime"],
            "DET.SEQ.TRIGGER": "T"
            if nodpattern
            else "F",  # wait for trigger if nodding
            "DET.TDELAY.GUI": 1000 * app_data["dwell"],
        }

        # parameters for user-defined headers
        user_data = setup_data.get("user", {})
        imagetyp = user_data.get("flags", "")

        # put OBSTYPE in for GTC. Also, correctly set object
        obstype = "Imaging"
        obj = user_data.get("target", "")
        if imagetyp == "bias":
            obj = "Bias"
            obstype = "Bias"
        elif imagetyp == "dark":
            obj = "Dark"
            obstype = "Dark"
        elif imagetyp == "flat":
            obj = "Skyflat"
            obstype = "SkyFlat"

        userpars = []

        # gtc headers
        telpars = setup_data.get("gtc_headers", {})
        try:
            gtc_header_info = create_header_from_telpars(telpars)
        except:
            gtc_header_info = None

        if gtc_header_info:
            userpars.extend([(item, gtc_header_info[item]) for item in gtc_header_info])
            gtcprgid = user_data.get("ID", "")

            gtcobid = user_data.get("OB", "1")
            # force 4 digit formatting
            try:
                gtcobid = "{:04d}".format(int(gtcobid))
            except ValueError:
                # not an integer
                pass

            # force for calibrations
            if obj in ("Skyflat", "Bias", "Dark"):
                gtcprgid = "Calib"

            # now force GTCOBID to be 'CALIB' if PRGID is
            # p.s this is a request from GTC staff
            if gtcprgid.lower().startswith("calib"):
                gtcobid = "CALIB"
            # add prgid and obid to headers
            userpars.extend([("GTCPRGID", gtcprgid), ("GTCOBID", gtcobid)])

        # now update with GUI values. Do this after to allow override of
        # GTC telescope server headers
        userpars.extend(
            [
                ("OBSERVER", user_data.get("Observers", "")),
                ("OBJECT", obj),
                ("RUNCOM", user_data.get("comment", "")),
                ("IMAGETYP", imagetyp),
                ("OBSMODE", obstype),
                ("FILTERS", user_data.get("filters", "")),
                ("PROGRM", user_data.get("ID", "")),
                ("PI", user_data.get("PI", "")),
            ]
        )

        # data from TCS
        tcs_data = setup_data.get("tcs", {})
        userpars.extend(
            [
                ("TELESCOP", tcs_data.get("tel", "WHT")),
                ("RA", tcs_data.get("RA", "00:00:00.00")),
                ("DEC", tcs_data.get("DEC", "+00:00:00.0")),
                ("ELEVAT", tcs_data.get("alt", -999)),
                ("AZIMUTH", tcs_data.get("az", -999)),
                ("AIRMASS", tcs_data.get("secz", -999)),
                ("INSTRPA", tcs_data.get("pa", -999)),
                ("TELFOCUS", tcs_data.get("foc", -999)),
                ("MOONDIST", tcs_data.get("mdist", -999)),
            ]
        )
        if tcs_data.get("tel", "WHT") == "GTC":
            userpars.extend([("ORIGIN", "GRANTECAN")])
        else:
            userpars.extend([("ORIGIN", "ING")])

        # data from h/w monitoring processes
        hw_data = setup_data.get("hardware", {})
        userpars.extend(
            [
                ("CCD1TEMP", hw_data.get("ccd1temp", -999)),
                ("CCD2TEMP", hw_data.get("ccd2temp", -999)),
                ("CCD3TEMP", hw_data.get("ccd3temp", -999)),
                ("CCD4TEMP", hw_data.get("ccd4temp", -999)),
                ("CCD5TEMP", hw_data.get("ccd5temp", -999)),
                ("CCD1VAC", hw_data.get("ccd1vac", -999)),
                ("CCD2VAC", hw_data.get("ccd2vac", -999)),
                ("CCD3VAC", hw_data.get("ccd3vac", -999)),
                ("CCD4VAC", hw_data.get("ccd4vac", -999)),
                ("CCD5VAC", hw_data.get("ccd5vac", -999)),
                ("CCD1FLOW", hw_data.get("ccd1flow", -999)),
                ("CCD2FLOW", hw_data.get("ccd2flow", -999)),
                ("CCD3FLOW", hw_data.get("ccd3flow", -999)),
                ("CCD4FLOW", hw_data.get("ccd4flow", -999)),
                ("CCD5FLOW", hw_data.get("ccd5flow", -999)),
                ("FPSLIDE", hw_data.get("fpslide", -999)),
            ]
        )

        # data from COMPO
        compo_data = setup_data.get("compo", {})
        userpars.extend(
            [
                ("COMPOPOA", compo_data.get("pickoff_angle", -999)),
                ("COMPOINJ", compo_data.get("injection_side", "UNDEF")),
            ]
        )
        # getting pars from GTC header info and TCS/GUI can insert duplicates
        # use a dict to remove duplicates
        self.userpars = OrderedDict(userpars)

    def setup_acq_task(self, nq=5):
        gps = 1 if self.use_gps_hardware else 0
        self.acq_dict = dict(nq=nq, gps=gps)
        if self.readoutMode == 4:
            # drift mode
            self.acq_dict["drf"] = 1
        else:
            self.acq_dict["drf"] = 0

        if self.detpars["DET.SPEED"] == 0:
            # slow
            self.acq_dict["aveg"] = 4
        else:
            # fast
            self.acq_dict["aveg"] = 1

    @property
    def acq_command(self):
        template = "acqproc hiperCamCCD -chip 5 -t 4 -nq {nq} -deml 0 -gps {gps} -aveg {aveg} -aps 8 -drf {drf}"
        return template.format(**self.acq_dict)

    @property
    def readmode_command(self):
        return "setup DET.READ.CURID {}".format(self.readoutMode)

    @property
    def setup_command(self):
        setup_string = "setup"

        for key in self.detpars:
            setup_string += " {} {} ".format(key, self.detpars[key])

        if self.finite:
            setup_string += " DET.FRAM2.BREAK {} ".format(self.finite)

        return setup_string

    @property
    def header_commands(self):
        def chunks(data, SIZE=10):
            it = iter(data)
            for i in range(0, len(data), SIZE):
                yield {k: data[k] for k in islice(it, SIZE)}

        setup_strings = []
        for chunk in chunks(self.userpars):
            setup_string = "setup"
            for key in chunk:
                if chunk[key] != "":
                    value = chunk[key]
                    # add quotes to strings with spaces
                    try:
                        if " " in value:
                            value = '"' + value + '"'
                    except:
                        pass
                    setup_string += " {} {} ".format(key, value)
            setup_strings.append(setup_string)
        return setup_strings


class FullFrame(ObsMode):
    def __init__(self, setup_data):
        super(FullFrame, self).__init__(setup_data)
        self.readoutMode = 1
        self.setup_acq_task(nq=5)


class Idle(ObsMode):
    """
    An idle mode to keep clearing the chip and stop taking GPS timestamps.
    """

    def __init__(self):
        app_data = {
            "xbin": 1,
            "ybin": 1,
            "clear": True,
            "readout": "Slow",
            "led_flsh": False,
            "oscan": False,
            "dummy": 1,
            "numexp": 0,
            "oscany": False,
            "multipliers": (1, 1, 1, 1, 1),
            "exptime": 10,
            "dwell": 10,
        }
        setup_data = {"appdata": app_data}
        super(Idle, self).__init__(setup_data)
        self.detpars["DET.GPS"] = "F"
        self.readoutMode = 1  # FF, Slow
        self.setup_acq_task(nq=5)


class Windows(ObsMode):
    def __init__(self, setup_data):
        super(Windows, self).__init__(setup_data)
        app_data = setup_data["appdata"]
        win1 = {
            "DET.WIN1.NX": app_data["x1size"],
            "DET.WIN1.NY": app_data["y1size"],
            "DET.WIN1.XSE": app_data["x1start_lowerleft"] - 1,
            "DET.WIN1.XSF": 2049 - app_data["x1start_lowerright"] - app_data["x1size"],
            "DET.WIN1.XSG": 2049 - app_data["x1start_upperright"] - app_data["x1size"],
            "DET.WIN1.XSH": app_data["x1start_upperleft"] - 1,
            "DET.WIN1.YS": app_data["y1start"] - 1,
            "DET.WIN1.YS.GUI": app_data["y1start"],
            "DET.WIN1.XSLL": app_data["x1start_lowerleft"],
            "DET.WIN1.XSLR": app_data["x1start_lowerright"],
            "DET.WIN1.XSUL": app_data["x1start_upperleft"],
            "DET.WIN1.XSUR": app_data["x1start_upperright"],
        }
        self.detpars.update(win1)

        if "x2size" in app_data:
            win2 = {
                "DET.WIN2.NX": app_data["x2size"],
                "DET.WIN2.NY": app_data["y2size"],
                "DET.WIN2.XSE": app_data["x2start_lowerleft"] - 1,
                "DET.WIN2.XSF": 2049
                - app_data["x2start_lowerright"]
                - app_data["x2size"],
                "DET.WIN2.XSG": 2049
                - app_data["x2start_upperright"]
                - app_data["x2size"],
                "DET.WIN2.XSH": app_data["x2start_upperleft"] - 1,
                "DET.WIN2.YS": app_data["y2start"] - 1,
                "DET.WIN2.YS.GUI": app_data["y2start"],
                "DET.WIN2.XSLL": app_data["x2start_lowerleft"],
                "DET.WIN2.XSLR": app_data["x2start_lowerright"],
                "DET.WIN2.XSUL": app_data["x2start_upperleft"],
                "DET.WIN2.XSUR": app_data["x2start_upperright"],
            }
            self.detpars.update(win2)

        if self.readoutMode == 2:
            self.setup_acq_task(nq=30)
        else:
            self.setup_acq_task(nq=10)

    @property
    def readoutMode(self):
        if "DET.WIN2.NX" in self.detpars:
            mode = 3
        else:
            mode = 2
        return mode


class Drift(ObsMode):
    def __init__(self, setup_data):
        super(Drift, self).__init__(setup_data)
        app_data = setup_data["appdata"]
        win1 = {
            "DET.DRWIN.NX": app_data["x1size"],
            "DET.DRWIN.NY": app_data["y1size"],
            "DET.DRWIN.YS": app_data["y1start"] - 1,
            "DET.DRWIN.YS.GUI": app_data["y1start"],
            "DET.DRWIN.XSE": app_data["x1start_left"] - 1,
            "DET.DRWIN.XSF": 2049 - app_data["x1start_right"] - app_data["x1size"],
            "DET.DRWIN.XSH": app_data["x1start_left"] - 1,
            "DET.DRWIN.XSG": 2049 - app_data["x1start_right"] - app_data["x1size"],
            "DET.DRWIN.XSL": app_data["x1start_left"],
            "DET.DRWIN.XSR": app_data["x1start_right"],
        }
        self.detpars.update(win1)
        self.nrows = 520  # number of rows in storage area
        self.readoutMode = 4
        self.setup_acq_task(nq=800)

    @property
    def num_stacked(self):
        """
        Number of windows stacked in frame transfer area
        """
        ny = self.detpars["DET.DRWIN.NY"]
        return int((self.nrows / ny + 1) / 2)

    @property
    def pipe_shift(self):
        """
        Extra shift to add to some windows to ensure uniform exposure time.

        Returned in units of vertical clocks. Should be multiplied by the
        vclock time to obtain pipe_shift in seconds.
        """
        ny = self.detpars["DET.DRWIN.NY"]
        return self.nrows - (2 * self.num_stacked - 1) * ny



import time


class AxisEmulator(object):
    """
    Simple class to emulate a moving stage
    """
    def __init__(self, speed):
        # units of microsteps
        self.home_position = 0
        self._current_position = 0
        self._target_position = 0
        self.moving = False
        self.speed = speed

    @property
    def current_position(self):
        if not self.moving:
            return self._current_position

        deltaT = time.time() - self._time_check
        value = abs(int(deltaT * self.speed))

        if value > abs(self._target_position - self._current_position):
            value = abs(self._target_position - self._current_position)
            self.moving = False

        if self._current_position > self._target_position:
            value *= -1

        self._current_position += value
        self._time_check = time.time()
        return int(self._current_position)

    @property
    def target_position(self):
        return self._target_position

    @target_position.setter
    def target_position(self, value):
        if value != self.current_position:
            self._target_position = int(value)
            self._time_check = time.time()
            self.moving = True



"""
Low level interface to NGC controller sending messages over
TCP/IP ports using ngcbCmd and reading instance database.
"""
import subprocess
import re
import os
import time
import glob
import sqlite3  # only needed for testing


class NGCDevice(object):
    """
    Class to wrap low-level communication with the NGC control software
    """
    def send_command(self, command, *command_pars):
        """
        Use low level ngcbCmd to send command to control server.
        """
        command_array = ['ngcbCmd', command]
        if command_pars:
            command_array.extend(command_pars)
        # must contain only strings
        command_array = [str(val) for val in command_array]

        # now attempt to send command to control server
        try:
            ret_msg = subprocess.check_output(command_array).strip().decode()
            # make sure all successful return true as status
            result = (ret_msg, True)
        except subprocess.CalledProcessError as err:
            # make sure all failed commands end with NOK line
            result = (err.output.strip().decode(), False)
        return result

    def stop_ngc_sw(self):
        """
        Shuts down ESO server
        """
        output = {}
        cmd = ['ngcdcsStopServer', 'IRCAM1']
        result = subprocess.check_output(cmd).strip().decode()
        output[' '.join(cmd)] = result
        return output

    def start_ngc_sw(self, gui=False):
        """
        Setup NGC controller for use with HiPERCAM
        """
        output = {}
        port = str(os.getenv('NGC_PORT', 8080))
        if gui:
            cmd = ['ngcdcsStartServer', 'IRCAM1', '-port', port, '-gui']
        else:
            cmd = ['ngcdcsStartServer', 'IRCAM1', '-port', port]
        result = subprocess.check_output(cmd).strip().decode()
        output[' '.join(cmd)] = result

        return output

    def setup_ngc_hipercam(self):
        """
        To be run when device moves to standby state
        """
        output = {}
        # get last file name
        runs = glob.glob('/data/run*.fits')
        if len(runs) == 0:
            last = 'run0000.fits'
        else:
            last = sorted(runs)[-1]
        cmds = [
            'ngcbCmd setup DET.SYSCFG hipercam_5ccd.cfg',
            'ngcbCmd setup DET.FRAM.NAMING auto',
            'ngcbCmd setup DET.FRAM.FILENAME run',
            'dbWrite "<alias>ngcircon_ircam1:exposure.newDataFileName" {}'.format(last)
        ]
        for cmd in cmds:
            cmd = cmd.split()
            result = subprocess.check_output(cmd).strip().decode()
            time.sleep(0.05)
            output[' '.join(cmd)] = result
            time.sleep(0.05)

        return output

    @property
    def database_summary(self):
        """
        Get a summary of system state and exposure state from database.
        """
        cmdTemplate = 'dbRead "<alias>ngcircon_ircam1:{}"'
        database_attributes = [
            'system.stateName',
            'system.subStateName',
            'cldc_0.statusName',
            'seq_0.statusName',
            'exposure.time',  # total exposure time for run
            'exposure.countDown',  # time remaining
            'exposure.expStatusName',
            'exposure.baseName',
            'exposure.newDataFileName']  # LAST WRITTEN FILE
        results = {}
        for attribute in database_attributes:
            cmd = cmdTemplate.format(attribute)
            try:
                response = subprocess.check_output(cmd, shell=True)
                results[attribute] = response.decode().split('=')[-1].strip()
            except Exception:
                results[attribute] = 'ERR'
        return results


class NGCEmulator(object):

    known_params = {
        'DET.NDIT': 10,
        'DET.SEQ1.DIT': 5,
        'DET.SEQ2.DIT': 5,
        'DET.SEQ3.DIT': 5,
        'DET.SEQ4.DIT': 5,
        'DET.SEQ5.DIT': 5,
        'DET.FRAM2.NO': 1
    }
    database = ".test_ngc_server.db"

    def __init__(self):
        self.past_time = 1400000000
        self.create_db()

    def send_command(self, command, *command_pars):
        known_commands = {
            'start': self.start,
            'abort': self.abort,
            'stop': self.stop,
            'online': self.online,
            'off': self.off,
            'standby': self.standby,
            'reset': self.reset,
            'status': self.status,
            'acqproc': self.acqproc,
            'seq 0 start': self.seqstart,
            'seq 0 stop': self.seqstop,
            'seq start': self.seqstart,
            'seq stop': self.seqstop,
            'cldc 0 enable': self.pon,
            'cldc 0 disable': self.poff,
            'trigger': self.trigger,
            'exit': self.exit,
            'setup': self.setup
        }
        if command not in known_commands:
            commands = command.split()
            command = commands[0]
            args = commands[1:]
            command_pars = list(command_pars)
            command_pars.extend(args)
            if command not in known_commands:
                raise ValueError('unknown command: ' + command)

        cmd = known_commands[command]
        result, ok = cmd(*command_pars)
        return result, ok

    def acqproc(self, *args):
        print('setting acquisition process', args)

    def start(self, *args):
        # start run
        self.query_db("update run set startTime = ? WHERE id = 1", time.time())
        self.query_db("update run set expState = ? WHERE id = 1", "active")
        return "OK", True

    def pon(self, *args):
        self.query_db("update run set clocks = ? WHERE id = 1", 'enabled')
        return "OK", True

    def poff(self, *args):
        self.query_db("update run set clocks = ? WHERE id = 1", 'disabled')
        return "OK", True

    def seqstart(self, *args):
        self.query_db("update run set sequencer = ? WHERE id = 1", 'running')
        return "OK", True

    def seqstop(self, *args):
        self.query_db("update run set sequencer = ? WHERE id = 1", 'idle')
        return "OK", True

    def abort(self, *args):
        # abort run
        # fake stopping run by setting starttime a long time in the past
        self.query_db("update run set startTime = ? WHERE id = 1", self.past_time)
        return "OK", True

    def stop(self, *args):
        # stop run
        # fake stopping run by setting starttime a long time in the past
        self.query_db("update run set startTime = ? WHERE id = 1", self.past_time)
        # add 1 to run number
        row = self.query_db("select * from run", one=True)
        run_number = int(row['run'])
        run_number += 1
        self.query_db("update run set run = ? WHERE id = 1", run_number)
        return "OK", True

    def online(self, *args):
        # bring sw online
        self.query_db("update run set stateName = ? WHERE id = 1", 'ONLINE')
        return "OK", True

    def off(self, *args):
        # move to loaded state
        self.query_db("update run set stateName = ? WHERE id = 1", 'LOADED')
        return "OK", True

    def exit(self, *args):
        # move to off state
        self.query_db("update run set stateName = ? WHERE id = 1", 'OFF')
        return "OK", True

    def standby(self, *args):
        # standby state
        self.query_db("update run set stateName = ? WHERE id = 1", 'STANDBY')
        return "OK", True

    def reset(self, *args):
        # reset controller
        self.create_db()
        return "OK", True

    def trigger(self, *args):
        return "OK", True

    def status(self, *args):
        # get status. either of whole system or requested param
        if args:
            param = args[0]
            if param not in self.known_params:
                raise ValueError('param {} not in known parameters: '.format(param))
            val = self.known_params[param]
        else:
            row = self.query_db("select * from run", one=True)
            val = row['stateName']
        return val, True

    def setup(self, *parval_list):
        if len(parval_list) % 2 != 0:
            raise ValueError('uneven pairing of params and values')
        print('setup', ' '.join(parval_list))
        for param, value in zip(parval_list[::2], parval_list[1::2]):
            # set param to value
            if re.match('DET\.SEQ\d\.DIT', param):
                self.known_params[param] = value
                self.query_db("update run set DIT = ? WHERE id = 1", value)
            elif param == 'NDIT':
                self.known_params['NDIT'] = value
                self.query_db("update run set NDIT = ? WHERE id = 1", value)
            else:
                self.known_params[param] = value
        return "OK", True

    def start_ngc_sw(self, gui=False):
        print('starting NGC Software')
        self.query_db("update run set stateName = ? WHERE id = 1", 'LOADED')
        return {}

    def setup_ngc_hipercam(self):
        print('Initialising HiPERCAM specific settings')
        return {}

    @property
    def database_summary(self):
        row = self.query_db("select * from run", one=True)
        elapsedTime = time.time() - row['startTime']
        TOTAL = float(row['DIT']) * float(row['NDIT'])
        countDown = TOTAL - float(elapsedTime)
        run = int(row['run'])
        if elapsedTime > TOTAL:
            expStatusName = "success"
            subState = "idle"
            if row['expState'] != 'idle':
                self.query_db("update run set run = ? WHERE id = 1", run+1)
                self.query_db("update run set expState = ? WHERE id = 1", "idle")
        else:
            expStatusName = "integrating"
            subState = "active"

        summary_dictionary = {
            "exposure.baseName": "/data",
            "exposure.countDown": str(countDown),
            "exposure.expStatusName": expStatusName,
            "exposure.newDataFileName": "run{:04d}.fits".format(run),
            "exposure.time": "6",
            "cldc_0.statusName": row['clocks'],
            'seq_0.statusName': row['sequencer'],
            "system.stateName": row['stateName'],
            "system.subStateName": subState
        }
        return summary_dictionary

    def create_db(self):
        if os.path.exists(self.database):
            os.unlink(self.database)
        with sqlite3.connect(self.database) as conn:
            conn.execute(
                'CREATE TABLE run (id INT, startTime FLOAT, clocks TEXT, sequencer TEXT' +
                ', NDIT INT, DIT FLOAT, stateName TEXT, run INT, expState TEXT)'
            )
            keys = ('DET.SEQ{}.DIT'.format(seq+1) for seq in range(5))
            DIT = max([self.known_params[key] for key in keys])
            conn.execute(
                "INSERT INTO run (id, startTime, clocks, sequencer, NDIT, DIT, stateName, run, expState) " +
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (1, self.past_time, 'disabled', 'idle', self.known_params['DET.NDIT'], DIT, 'OFFLINE', 1, "idle")
            )
            conn.commit()

    def get_db(self):
        db = sqlite3.connect(self.database)
        db.row_factory = sqlite3.Row
        return db

    def close_db(self, db):
        db.close()

    def query_db(self, query, *args, **kwargs):
        one = kwargs.get('one', False)
        conn = self.get_db()
        try:
            cur = conn.execute(query, args)
            rv = cur.fetchall()
            conn.commit()
            cur.close()
        finally:
            self.close_db(conn)
        return (rv[0] if rv else None) if one else rv



# talk to honeywell temperature monitor
from __future__ import absolute_import, unicode_literals, print_function, division
import threading
import six
from random import normalvariate

try:
    from pymodbus.constants import Endian
    from pymodbus.payload import BinaryPayloadDecoder
    from pymodbus.client import ModbusTcpClient as ModbusClient
except:
    from pymodbus3.constants import Endian
    from pymodbus3.payload import BinaryPayloadDecoder
    from pymodbus3.client.sync import ModbusTcpClient as ModbusClient    

class MinitrendError(Exception):
    pass


class Minitrend:
    def __init__(self, address, port):
        self.address = address
        self.client = ModbusClient(address, port=port)
        # list mapping pen ID number to address
        self.pen_addresses = {i: 0x18C0 + i * 0x2 for i in range(16)}
        self.alarm_addresses = {i: 0x1980 + i * 0x1 for i in range(16)}
        self.unit_id = 0x01  # allows us to address different units on the same network
        # thread safe access
        self._lock = threading.Lock()

    def __del__(self):
        self.disconnect()

    def disconnect(self):
        self.client.close()

    def connect(self):
        success = self.client.connect()
        return success

    def read_register(self, address, size):
        """
        Get a payload from a register

        Raises
        ------
        MinitrendError
            When reading fails
        """
        with self._lock:
            try:
                payload = self.client.read_input_registers(
                    address, size, unit=self.unit_id
                )
            except Exception as err:
                raise MinitrendError(str(err))
        return payload

    def get_pen(self, pen_number):
        address = self.pen_addresses[pen_number]
        result = self.read_register(address, 2)
        if not six.PY3:
            decoder = BinaryPayloadDecoder.fromRegisters(
                result.registers, endian=Endian.Big
            )
        else:
            decoder = BinaryPayloadDecoder.from_registers(
                result.registers, endian=Endian.Big
            )
        return decoder.decode_32bit_float()

    def get_alarm(self, alarm_number):
        address = self.alarm_addresses[alarm_number]
        result = self.read_register(address, 1)
        if not six.PY3:
            decoder = BinaryPayloadDecoder.fromRegisters(
                result.registers, endian=Endian.Big
            )
        else:
            decoder = BinaryPayloadDecoder.from_registers(
                result.registers, endian=Endian.Big
            )
        value = decoder.decode_16bit_uint()
        # six alarms, each corresponding to a bit of the return value
        flags = [(value >> i) & 1 for i in range(6)]
        return flags


class MinitrendEmulator:
    def __init__(self, pen_values, noise_level=0):
        self.pen_values = pen_values
        self.noise_level = noise_level

    def connect(self):
        return True

    def disconnect(self):
        return

    def get_pen(self, pen_number):
        return self.pen_values + self.noise_level * normalvariate(0, 1)

    def get_alarm(self, alarm_number):
        return [0, 0, 0, 0, 0, 0]



# A serial class emulator. Designed to be an abstract class
class FakeSerial:
    simple_commands = {}
    registered_commands = {}

    def __init__(self, port, baudrate=19200, timeout=1,
                 bytesize=8, parity='N', stopbits=1, xonxoff=0,
                 rtscts=0):
        self.name = port
        self.port = port
        self.timeout = timeout
        self.parity = parity
        self.baudrate = baudrate
        self.bytesize = bytesize
        self.stopbits = stopbits
        self.xonxoff = xonxoff
        self.rtscts = rtscts
        self._isOpen = True
        self._receivedData = ""
        self._data = b""

    @classmethod
    def register_command(cls, name, type_conv):
        def inner(func):
            cls.registered_commands[name] = (func, type_conv)
            return func
        return inner

    def isOpen(self):
        return self._isOpen

    def open(self):
        self._isOpen = True

    def close(self):
        self._isOpen = False

    def write(self, string):
        self._receivedData += string.decode()
        self._process_received()

    def read(self, n=1):
        """
        reads n characters from data buffer
        """
        s = self._data[0:n]
        self._data = self._data[n:]
        return s.encode()

    def readline(self):
        try:
            returnIndex = self._data.index("\n")
        except ValueError:
            return ""
        else:
            s = self._data[0:returnIndex+1]
            self._data = self._data[returnIndex+1:]
            return s.encode()

    def __str__(self):
        template = "Serial<id=0xa81c10, open={}>(port='{}', baudrate={},"
        template += " bytesize={:d}, parity='{}', stopbits={:d}, xonxoff={:d}, rtscts={:d})"
        return template.format(
            str(self.isOpen()), self.port, self.baudrate,
            self.bytesize, self.parity, self.stopbits, self.xonxoff,
            self.rtscts
        )

    def _process_received(self):
        # if it's a simple command, send response
        if self._receivedData in self.simple_commands:
            cmd = self._receivedData
            self._receivedData = ""
            template = str(self.simple_commands[cmd])
            msg = template.format(**self.__dict__)
            self._data = msg + "\n"

        # for more complex commands, strip command from data
        # call function, append response
        for cmd in self.registered_commands:
            if self._receivedData.startswith(cmd):
                func, converter = self.registered_commands[cmd]

                arg = self._receivedData.strip(cmd)
                if len(arg) == 0:
                    result = func(self)
                else:
                    if converter is not None:
                        arg = converter(arg)
                    result = func(self, arg)

                result = str(result) + "\n" if result is not None else ""
                self._data = result
                self._receivedData = ""



# Utility to communicate with MKS PDR900 Vacuum Guage via RS232
from __future__ import absolute_import, unicode_literals, print_function, division
import re
import socket
import threading
import six

from astropy.utils.decorators import lazyproperty
from astropy.time import Time, TimeDelta
from astropy.io import ascii

from ..testing.sensors import SensorEmulator

DEFAULT_TIMEOUT = 5  # seconds

# MESSAGES (as strings, don't forget to format and encode)
DOWNLOAD = '@{addr:03d}DL{comm};FF'
SERIAL_NO = '@{addr:03d}SNC{comm};FF'
FIRMWARE = '@{addr:03d}FVC{comm};FF'
ADDRESS = '@{addr:03d}ADC{comm};FF'
DLOG_CTRL = '@{addr:03d}DLC{comm};FF'
DLOG_TIME = '@{addr:03d}DLT{comm};FF'
PRESSURE = '@{addr:03d}PR1{comm};FF'

# rename socket.error
if six.PY2:
    ConnectionRefusedError = socket.error


class VacuumGaugeError(Exception):
    pass


class PDR900Emulator(object):

    def __init__(self, host, port):
        self._sensor = SensorEmulator(1.0e-5, 1.0e-6, 0.0)

    def disconnect(self):
        pass

    def connect(self):
        pass

    @property
    def pressure(self):
        return self._sensor.current_value


class PDR900(object):

    def __init__(self, host, port):
        """
        Creates a PDR900 object for communication over serial.

        Parameters
        -----------
         port : string
            port device representing the vacuum gauge
        """
        self.port = port
        self.host = host
        self.logging_start_time = None
        try:
            self.connect()
        except ConnectionRefusedError:
            self.sock = None
        self._lock = threading.Lock()

    def __del__(self):
        self.close()

    def connect(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.host, self.port))
        s.settimeout(5)
        self.sock = s

    def close(self):
        if hasattr(self, 'sock'):
            self.sock.close()
        self.sock = None

    def _parse_response(self, response):
        pattern = '@(.*)ACK(.*);FF'
        result = re.match(pattern, response)
        if result is None:
            raise VacuumGaugeError('could not parse response {}'.format(
                response
            ))
        if len(result.groups()) != 2:
            raise VacuumGaugeError('unexpected result when parsing response {}'.format(
                response
            ))
        return result.groups()

    def _send_recv(self, message, data):
        # try and connect if we are not already
        msg = message.format(**data).encode()
        if self.sock is None:
            try:
                self.connect()
            except ConnectionRefusedError:
                raise VacuumGaugeError('not connected')

        self.sock.send(msg)
        self.sock.settimeout(DEFAULT_TIMEOUT)
        response = self.sock.recv(1024).decode().rstrip('\r\n')
        addr, retval = self._parse_response(response)
        return addr, retval

    @lazyproperty
    def address(self):
        # connect once to find address and hard-code
        data = dict(addr=254, comm='?')
        _, addr = self._send_recv(ADDRESS, data)
        return int(addr)

    @lazyproperty
    def firmware_version(self):
        data = dict(addr=self.address, comm='?')
        _, fwver = self._send_recv(FIRMWARE, data)
        return fwver

    @lazyproperty
    def serial_number(self):
        data = dict(addr=self.address, comm='?')
        _, serno = self._send_recv(SERIAL_NO, data)
        return serno

    @property
    def pressure(self):
        data = dict(addr=self.address, comm='?')
        addr, response = self._send_recv(PRESSURE, data)
        return float(response)

    def start_logging(self):
        data = dict(addr=self.address, comm='!START')
        addr, response = self._send_recv(DLOG_CTRL, data)
        if response != 'START':
            raise VacuumGaugeError('failed to start logging')
        self.logging_start_time = Time.now()

    def stop_logging(self):
        data = dict(addr=self.address, comm='!STOP')
        addr, response = self._send_recv(DLOG_CTRL, data)
        if response != 'STOP':
            raise VacuumGaugeError('failed to stop logging')
        self.logging_start_time = Time.now()

    def set_log_interval(self, hours, mins, secs):
        assert secs < 60, 'seconds must be less than 60'
        assert mins < 60, 'minutes must be less than 60'
        tstring = '!{:02d}:{:02d}:{:02d}'.format(hours, mins, secs)
        data = dict(addr=self.address, comm=tstring)
        addr, response = self._send_recv(DLOG_TIME, data)
        if response != tstring[1:]:
            raise VacuumGaugeError('failed to set logging time')

    def get_log_interval(self):
        data = dict(addr=self.address, comm='?')
        addr, response = self._send_recv(DLOG_TIME, data)
        try:
            h, m, s = [float(val) for val in response.split(':')]
        except Exception:
            raise VacuumGaugeError('cannot parse log interval response: ' + response)
        return TimeDelta(3600*h + 60*m + s, format='sec')

    def get_log_data(self):
        data = dict(addr=self.address, comm='?')
        addr, pdata = self._send_recv(DOWNLOAD, data)
        pdata = pdata.rstrip('\x03').replace('\r', '\n')
        return ascii.read(pdata, delimiter=';')



import requests
import json
import warnings
import os

SERVER = os.getenv('WAMP_SERVER', "192.168.1.2")
BASE_URL = "http://{}:8080/".format(SERVER)
CALL_URL = BASE_URL + "call"
PUBLISH_URL = BASE_URL + "publish"


class WAMPRuntimeError(Exception):
    pass


def _send(url, destination, *args, **kwargs):
    json_key = 'procedure' if url == CALL_URL else 'topic'
    json = {json_key: destination}
    if args:
        json['args'] = args
    if kwargs:
        json['kwargs'] = kwargs
    response = requests.post(url, json=json)
    response.raise_for_status()
    return response.content.decode()


def _check_procedure(procedure):
    """
    Check if remote procedures registered with server
    """
    data = json.loads(_send(CALL_URL, 'wamp.registration.lookup', procedure))
    return bool(data['args'][0])


def _check_topic(topic):
    """
    Check if anyone has subscribed to this topic on server
    """
    data = json.loads(_send(CALL_URL, 'wamp.subscription.lookup', topic))
    return bool(data['args'][0])


def call(procedure, *args, **kwargs):
    """
    Call a remote procedure that has been registered with the server

    Parameters
    ----------
    procedure : str
        name of procedure, e.g 'hipercam.Compo.rpc.power_off_axis'
    *args : (optional) variable length argument list
        arguments to pass to remote procedure
    **kwargs : (optional)
        dictionary of kwargs to pass to remote procedure

    Returns
    -------
    result
        the result of calling the remote procedure
    """
    if not _check_procedure(procedure):
        raise ValueError('procedure "{}" not registered with WAMP server'.format(
            procedure))
    response = json.loads(_send(CALL_URL, procedure, *args, **kwargs))
    if "error" in response:
        raise WAMPRuntimeError(
            response['args']
        )
    return response['args'][0]


def publish(topic, *args, **kwargs):
    """
    Publish data to a topic

    The routine will warn, but not fail, if no-one is subscribed to the topic

    Parameters
    ----------
    topic : str
        name of topic, e.g 'hipercam.FocalPlaneSlide.target_position'
    *args : (optional) variable length argument list
        arguments to publish to topic
    **kwargs : (optional)
        dictionary of kwargs to publish to topic

    Returns
    -------
    result
        the response from the server
    """
    if not _check_topic(topic):
        warnings.warn(
            'no-one has subscribed to topic {}. your message may not be read'.format(
                topic
            )
        )

    response = json.loads(_send(PUBLISH_URL, topic, *args, **kwargs))
    msg_id = response['id']
    return msg_id



from typing import Any
from ..wamp import ModelComponentMixin, WrapperComponentMixin
from ..models.slide import FocalPlaneSlide
from ..models.flow_sensor import FlowSensor
from ..models.compo import Compo, CompoArms, CompoLens
from ..models.meerstetter import Meerstetter
from ..models.vac_gauge import VacGauge
from ..models.ccd import CCDHead
from ..models.ngc import NGC
from ..models.telescope import GTC
from ..models.rack import Rack


class RackComponent(ModelComponentMixin):
    def __init__(self, config):
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        super(RackComponent, self).__init__(config)
        self.traceback_app = True
        emulation_mode = config.extra.get("emulate")
        self.model = Rack(emulate=emulation_mode)


class GTCComponent(ModelComponentMixin):
    def __init__(self, config):
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        super(GTCComponent, self).__init__(config)
        self.traceback_app = True
        emulation_mode = config.extra.get("emulate")
        path = config.extra.get("path", "/data")
        self.model = GTC(emulate=emulation_mode, path=path)


class NGCComponent(ModelComponentMixin):
    def __init__(self, config):
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        super(NGCComponent, self).__init__(config)
        self.traceback_app = True
        emulation_mode = config.extra.get("emulate")
        self.model = NGC(emulation_mode)


class CCDComponent(WrapperComponentMixin):
    def __init__(self, config):
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        pressure_topic = config.extra.get("pressure_topic")
        flow_topic = "hipercam.flow"
        peltier_topic = config.extra.get("peltier_topic")
        pen_number = config.extra.get("pen_number")
        peltier_channel = config.extra.get("peltier_channel")

        super(CCDComponent, self).__init__(config)
        self.traceback_app = True
        self.model = CCDHead(
            pressure_topic, flow_topic, peltier_topic, pen_number, peltier_channel
        )


class PressureComponent(ModelComponentMixin):
    def __init__(self, config):
        super(PressureComponent, self).__init__(config)
        self.traceback_app = True
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        host = config.extra.get("host", None)
        port = config.extra.get("port", None)
        emulation_mode = config.extra.get("emulate")
        self.model = VacGauge(host, port, emulate=emulation_mode)


class FocalPlaneSlideComponent(ModelComponentMixin):
    def __init__(self, config):
        super(FocalPlaneSlideComponent, self).__init__(config)
        self.traceback_app = True
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        host = config.extra.get("host", None)
        port = config.extra.get("port", None)
        emulation_mode = config.extra.get("emulate")
        self.model = FocalPlaneSlide(host, port, emulate=emulation_mode)


class FlowSensorComponent(ModelComponentMixin):
    def __init__(self, config):
        super(FlowSensorComponent, self).__init__(config)
        self.traceback_app = True
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        host = config.extra.get("host", None)
        port = config.extra.get("port", None)
        emulation_mode = config.extra.get("emulate")
        self.model = FlowSensor(host, port, emulate=emulation_mode)


class CompoArmComponent(ModelComponentMixin):
    def __init__(self, config):
        super(CompoArmComponent, self).__init__(config)
        self.traceback_app = True
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        port = config.extra.get("port", None)
        emulation_mode = config.extra.get("emulate")
        self.model = CompoArms(port, emulate=emulation_mode)


class CompoLensComponent(ModelComponentMixin):
    def __init__(self, config):
        super(CompoLensComponent, self).__init__(config)
        self.traceback_app = True
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        host = config.extra.get("host", None)
        port = config.extra.get("port", None)
        emulation_mode = config.extra.get("emulate")
        self.model = CompoLens(host, port, emulate=emulation_mode)


class CompoComponent(WrapperComponentMixin):
    def __init__(self, config):
        # get name from config, or class name if missing
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        arm_topic = config.extra.get("arm_topic")
        lens_topic = config.extra.get("lens_topic")

        super(CompoComponent, self).__init__(config)
        self.traceback_app = True
        self.model = Compo(arm_topic, lens_topic)


class MeerstetterComponent(ModelComponentMixin):
    def __init__(self, config):
        super(MeerstetterComponent, self).__init__(config)
        self.traceback_app = True
        self._component_name = config.extra.get(
            "name", str(self.__class__).replace("Component", "")
        ).lower()
        host = config.extra.get("host", None)
        port = config.extra.get("port", None)
        emulation_mode = config.extra.get("emulate")
        self.model = Meerstetter(host, port, emulate=emulation_mode)



# Utility to query and set temp, and start circulation on Huber Unichiller 025-MPC+
from __future__ import absolute_import, unicode_literals, print_function, division
import time
import socket
import threading
import six


QUERY_DEV = '[M01V07'
QUERY_STATUS = '[M01G0D******'
STATE_CONTROL = '[M01G0D{}*****'
TEMP_SET = '[M01G0D**{}'
DEFAULT_TIMEOUT = 2

# rename socket.error
if six.PY2:
    ConnectionRefusedError = socket.error


def hex_to_float(hexstring):
    """
    The chiller encodes floats using the following hex convevntion.

    1. convert hex to int
    2. if int val > 32767, it's a negative number, subtract 65536
    3. divide by temp resolution (100)
    """
    val = int(hexstring, 16)
    if val > 32767:
        val -= 65536
    return val/100


def float_to_hex(float):
    # reverse of above
    intval = int(float*100)
    if intval <= 0:
        intval += 65536
    return '{:0>4X}'.format(intval)[-4:]


class UnichillerException(Exception):
    pass


class UnichillerMPC(object):
    """
    Class to use serial-over-ethernet to communicate with a 025-MPC
    Unichiller from Huber.

    See LAI protocol spec document for details.
    """
    def __init__(self, host, port):
        self.host = host
        self.port = port

        # thread lock for threadsafe operations
        self._lock = threading.Lock()

    def __del__(self):
        # must close connection on object deletion
        self.close()

    def connect(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.host, self.port))
        s.settimeout(5)
        self.sock = s

    def close(self):
        if self.sock is not None:
            self.sock.close()
        self.sock = None

    def _checksum(self, msg):
        """
        Add a checksum to message.

        The checksum is a single byte (2 hex characters) sum of all the hex
        values in the message.

        Parameters
        ----------
        msg : string, bytes
            the message to calculate checksum of
        """
        return '{:>02X}'.format(sum(bytearray(msg, 'ascii')))[-2:]

    def _send_recv(self, msg):
        msg += self._checksum(msg) + '\r'
        with self._lock:
            self.sock.send(msg.encode())
            response = self.sock.recv(1024).decode().rstrip('\r')
        self._check_response(msg, response)
        return response

    def _check_response(self, msg, response):
        if response[0:4] == '[S01':
            # a valid response so far so good
            cs = response[-2:]
            package = response[:-2]
            if self._checksum(package) != cs:
                raise IOError('checksum of return message not OK:\nOut: {}\nBack: {}'.format(
                                msg, response
                              ))
        else:
            raise IOError('response not from device, or malformed:\nOut: {}\nBack: {}'.format(
                            msg, response
                          ))

    def get_status(self):
        response = self._send_recv(QUERY_STATUS)
        body = response[7:-2]
        status = dict()
        op_status = body[0]
        status['mode'] = {
            'C': 'pump on',
            'I': 'pump on, cooling on',
            'O': 'control off'
        }[op_status]
        alarm_status = body[1]
        status['alarms'] = True if alarm_status == '1' else False
        status['setpoint'] = hex_to_float(body[2:6])
        status['chiller_temp'] = hex_to_float(body[6:10])
        return status

    @property
    def temperature(self):
        status = self.get_status()
        return status['chiller_temp']

    @temperature.setter
    def temperature(self, target):
        target_hex = float_to_hex(target)
        msg = TEMP_SET.format(target_hex)
        self._send_recv(msg)

    def pump_on(self):
        msg = STATE_CONTROL.format('C')
        self._send_recv(msg)
        time.sleep(2)
        msg = STATE_CONTROL.format('I')
        self._send_recv(msg)

    def pump_off(self):
        msg = STATE_CONTROL.format('O')
        self._send_recv(msg)



import pickle

from autobahn import wamp
from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.types import PublishOptions
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import LoopingCall
from twisted.internet import reactor
from twisted.logger import Logger
from twisted.internet.error import ReactorNotRunning


@wamp.error(u"com.hipercam.runtimeerror")
class HipercamRuntimeError(Exception):
    """
    An application specific exception that is decorated with a WAMP URI,
    and hence can be automapped by Autobahn.
    """
    pass


TELEMETRY_DELAY = 2
HEARTBEAT_DELAY = 300


class WrapperComponentMixin(ApplicationSession):
    """
    A mixin class for a WAMP component that wraps others.
    """
    log = Logger()

    def onUserError(self, fail, msg):
        self.log.error(msg)

    @inlineCallbacks
    def onJoin(self, details):

        self.telemetry_topic = 'hipercam.{}.telemetry'.format(self._component_name)

        self.log.info("{} joined WAMP session: realm={}, session={}".format(
            self._component_name, details.realm, details.session
        ))

        def make_pass_through(to_topic):
            def inner(*args):
                self.publish(to_topic, *args)
            return inner

        # pass through topics
        for name in self.model.pass_through_topics:
            topic = 'hipercam.{}.{}'.format(
                self._component_name, name
            )
            to_topic = self.model.pass_through_topics[name]
            yield self.subscribe(make_pass_through(to_topic), topic)
            self.log.info('{} is passing messages on {} to {}'.format(
                self._component_name, topic, to_topic
            ))

        # subscribe to all the telemetry channels
        for channel_id in self.model.subscribed_channels:
            channel = self.model.subscribed_channels[channel_id]
            topic = channel + '.telemetry'
            yield self.subscribe(self.model.on_received_telemetry, topic)
            self.log.info("{} subscribed to topic {}".format(
                self._component_name, topic)
            )

        # install a heartbeat logger
        self._tick_no = 0
        self._tick_loop = LoopingCall(self._tick)
        self._tick_loop.start(HEARTBEAT_DELAY)

        # begin publishing telemetry
        self._telemetry_publisher = LoopingCall(self._publish_telemetry)
        self._telemetry_publisher.start(TELEMETRY_DELAY)
        self.log.info("{} started publishing telemetry on topic {}".format(self._component_name, self.telemetry_topic))

    def _tick(self):
        self._tick_no += 1
        self.log.info('{} is alive [tick {}]'.format(
            self._component_name, self._tick_no))

    @inlineCallbacks
    def _publish_telemetry(self):
        pub_options = PublishOptions(
            acknowledge=True,
            exclude_me=True
        )
        package = pickle.dumps(self.model.telemetry())
        publication = yield self.publish(self.telemetry_topic, package, options=pub_options)
        self.log.debug("Published telemetry to {} with publication ID {}".format(
            self.telemetry_topic, publication.id))

    def onLeave(self, details):
        self.log.info("Session closed: {details}", details=details)
        self.disconnect()

    def onDisconnect(self):
        self.log.info("Connection closed")
        if hasattr(self, '_tick_loop') and self._tick_loop:
            self._tick_loop.stop()
            self._tick_loop = None
        if hasattr(self, '_telemetry_publisher') and self._telemetry_publisher:
            self._telemetry_publisher.stop()
            self._telemetry_publisher = None
        try:
            reactor.stop()
        except ReactorNotRunning:
            pass


class ModelComponentMixin(ApplicationSession):
    """
    A mixin class for a WAMP Component reflecting a single model.

    Telemetry is regularly published, and a tick is published regularly
    to check if the component is alive.

    In addition, the component subscribes to topics for each settable
    attribute of the model, and registers RPC calls for any defined
    methods on the model, and all state transitions of the model.
    """
    log = Logger()

    def onUserError(self, fail, msg):
        self.log.error(msg)

    @inlineCallbacks
    def onJoin(self, details):

        self.telemetry_topic = 'hipercam.{}.telemetry'.format(self._component_name)

        self.log.info("{} joined WAMP session: realm={}, session={}".format(
            self._component_name, details.realm, details.session
        ))

        # start a looping call step state machines on
        # needs to be on a tigher loop than poll times in the machines,
        # of order 1s. Since there are many machines in a model,
        # we work out the loop time here
        nmachines = len(self.model.machines)
        if nmachines > 0:
            loop_time = 0.3/nmachines
            self._machine_loop = LoopingCall(self._forward_step_statemachines)
            self._machine_loop.start(loop_time)

        # add any Looping calls defined in the model
        self._model_looping_calls = []
        if hasattr(self.model, 'extra_looping_calls'):
            for call in self.model.extra_looping_calls:
                interval = self.model.extra_looping_calls[call]
                function = getattr(self.model, call)
                loop = LoopingCall(function)
                self._model_looping_calls.append(loop)
                loop.start(interval)

        # register all triggers of state transistions as RPC callbacks
        def make_trigger(machine, event):
            # use this factory function to avoid statemachine
            # being returned from trigger (it cannot be pickled)
            def f(**event_parameters):
                allowed_events = machine.statechart.events_for(machine.configuration)
                if event not in allowed_events:
                    raise HipercamRuntimeError('unsupported state transition on {}\n{} not in {}'.format(
                                        self._component_name, event, ",".join(allowed_events)
                                    ))
                machine.queue(event, **event_parameters)
            return f

        for machine_name in self.model.machines:
            self.log.info('registering state transition triggers')
            machine = self.model.machines[machine_name]
            events = machine.statechart.events_for()
            self.log.info('registering {} for {}'.format(events, machine_name))
            for event in events:
                rpc_name = 'hipercam.{}.rpc.{}.{}'.format(
                    self._component_name, machine_name, event
                )
                trigger = make_trigger(machine, event)
                yield self.register(trigger, rpc_name)
                self.log.info("{} registered procedure {}".format(self._component_name, rpc_name))

        # add additional RPC calls
        for method_name in self.model.extra_rpc_calls:
            trigger_function = getattr(self.model, method_name)
            rpc_name = 'hipercam.{}.rpc.{}'.format(
                self._component_name, method_name
            )
            yield self.register(trigger_function, rpc_name)
            self.log.info("{} registered procedure {}".format(self._component_name, rpc_name))

        # subscribe to channels corresponding to settable attributes

        # needs a factory function because of scopes
        # see https://eev.ee/blog/2011/04/24/gotcha-python-scoping-closures/
        def make_setter(attr):
            def setter(val):
                setattr(self.model, attr, val)
            return setter

        for i in range(len(self.model.settable_attributes)):
            # declared inside loop so it has proper scope for capture in callback
            attr = str(self.model.settable_attributes[i])
            topic = 'hipercam.{}.{}'.format(
                self._component_name, attr
            )
            yield self.subscribe(make_setter(attr), topic)
            self.log.info("{} registered settable attribute {} on topic {}".format(self._component_name, attr, topic))

        # install a heartbeat logger
        self._tick_no = 0
        self._tick_loop = LoopingCall(self._tick)
        self._tick_loop.start(HEARTBEAT_DELAY)

        # begin publishing telemetry
        self._telemetry_publisher = LoopingCall(self._publish_telemetry)
        self._telemetry_publisher.start(TELEMETRY_DELAY)
        self.log.info("{} started publishing telemetry on topic {}".format(self._component_name, self.telemetry_topic))

    def _forward_step_statemachines(self):
        """
        Step each state machine in the model forward.

        Needs to be on a tight control loop, because if this loops
        slower than poll times in the state machines, unusual
        behaviour can occur.

        This is because we want bound state changes between, e.g
        connection handlers and sensors, to propagate faster than
        the poll time, so devices don't try to poll a closed connection

        Each model can have many machines, and the fastest poll times
        are of order 1 second
        """
        for machine_name in self.model.machines:
            machine = self.model.machines[machine_name]
            machine.execute()

    def _tick(self):
        self._tick_no += 1
        self.log.info('{} is alive [tick {}]'.format(
            self._component_name, self._tick_no))

    @inlineCallbacks
    def _publish_telemetry(self):
        pub_options = PublishOptions(
            acknowledge=True,
            exclude_me=True
        )
        package = pickle.dumps(self.model.telemetry())
        publication = yield self.publish(self.telemetry_topic, package, options=pub_options)
        self.log.debug("Published telemetry to {} with publication ID {}".format(
            self.telemetry_topic, publication.id))

    def onLeave(self, details):
        self.log.info("Session closed: {details}", details=details)
        self.disconnect()

    def onDisconnect(self):
        self.log.info("Connection closed")
        if hasattr(self, '_model_looping_calls') and self._model_looping_calls:
            for call in self._model_looping_calls:
                call.stop()
            self._model_looping_calls = None
        if hasattr(self, '_tick_loop') and self._tick_loop:
            self._tick_loop.stop()
            self._tick_loop = None
        if hasattr(self, '_telemetry_publisher') and self._telemetry_publisher:
            self._telemetry_publisher.stop()
            self._telemetry_publisher = None
        if hasattr(self, '_machine_loop') and self._machine_loop:
            self._machine_loop.stop()
            self._machine_loop = None
        try:
            reactor.stop()
        except ReactorNotRunning:
            pass



import time


class SimulatedDeviceProperty(object):
    """
    A descriptor for access to a simulated property of a Device

    For example, one might wish to get or set the acceleration of a motor
    stage. There is a fair amount of boiler-plate code involved in getting
    or setting this property, but all that differs between them is the command
    code used and the units of the property.

    This is a perfect use of a Python descriptor
    """
    def __init__(self, name, initial, response_speed):
        """
        Parameters
        ----------
        initial : float
            initial value for property
        response_speed : float
            the rate at which property changes per second
        """
        self.name = name
        self._current_value = initial
        self._target_value = initial
        self._time_check = time.time()
        self._response_speed = response_speed

    def __get__(self, instance, cls):
        """
        In getter, we check when we last updated value and
        return the new value based on that
        """
        if instance is None:
            # called from class itself
            return self
        else:
            # called from class instance
            deltaT = time.time() - self._time_check
            value = deltaT * self._response_speed
            value = min(value, abs(self._target_value - self._current_value))
            if self._current_value - self._target_value > 0:
                value *= -1
            self._current_value += value
            self._time_check = time.time()
            return self._current_value

    def __set__(self, instance, value):
        # set value
        self._target_value = value
        # store the instance value
        instance.__dict__[self.name + '_target'] = value
        self._time_check = time.time()


class SimulatedIntegerDeviceProperty(SimulatedDeviceProperty):
    """
    An Device property that must be integer.
    """
    def __get__(self, instance, cls):
        value = super(SimulatedIntegerDeviceProperty, self).__get__(instance, cls)
        return int(value)

    def __set__(self, instance, value):
        super(SimulatedIntegerDeviceProperty, self).__set__(instance, int(value))


class SimulatedBooleanDeviceProperty(SimulatedDeviceProperty):
    """
    An Device property that must be integer.
    """
    def __get__(self, instance, cls):
        value = super(SimulatedBooleanDeviceProperty, self).__get__(instance, cls)
        return int(value)

    def __set__(self, instance, value):
        super(SimulatedBooleanDeviceProperty, self).__set__(instance, int(value))



import serial
import threading
import signal
import time
from functools import partial
from contextlib import contextmanager

from astropy import units as u

from .properties import DeviceProperty, microstep, IntegerDeviceProperty
from ..testing.fakeserial import FakeSerial
from ..testing.axis import AxisEmulator


class NoKeyboardInterrupt:
    """
    Context manager to ignore Keyboard interrupt for lifetime of context

    Interuppt is triggered upon exit. If called in a thread, does nothing.
    """
    def __enter__(self):
        try:
            self.signal_received = False
            self.old_handler = signal.getsignal(signal.SIGINT)
            signal.signal(signal.SIGINT, self.handler)
            self.threaded = False
        except ValueError:
            self.threaded = True

    def handler(self, signal, frame):
        self.signal_received = (signal, frame)

    def __exit__(self, type, value, traceback):
        if not self.threaded:
            signal.signal(signal.SIGINT, self.old_handler)
            if self.signal_received:
                self.old_handler(*self.signal_received)
        else:
            pass


class NewportError(Exception):
    '''
    Class to represent errors from the controller

    Errors come in the form (code, timestamp, MESSAGE).
    e.g 0, 451322, NO ERROR DETECTED
    '''
    def __init__(self, string):
        self._string = string
        code, ts, msg = string.split(',')
        if len(code) == 3:
            self.axis = int(code[0])
            self.code = int(code[1:])
        else:
            self.axis = None
            self.code = code
        self.timestamp = int(ts.strip())  # number of 400us intervals since reset
        self.message = msg.strip()

    def __str__(self):
        msg = 'newport error code {}'.format(self.code)
        if self.axis is not None:
            msg += ' on axis {}'.format(self.axis)
        return '{} ({})'.format(msg, self.message)


class NewportESP301:
    """
    Controller for ESP301 Motion Controller over Serial Interface.

    The controller itself has some limited functionality, but mostly this class serves
    as a container for a list of NewportESP301Axis objects, which directly control stages
    attached to each axis. For example

    >>> esp = NewportESP301('/dev/ttyUSB0')
    >>> esp.axis[0].home()

    Axes are numbered starting from zero, following Python standards, but in the controller
    they are labelled from 1, so `esp.axis[0]` refers to axis 1 as described in the manual.
    """
    def __init__(self, port, baudrate=19200, unit_equivalencies=None, emulate=False):
        """
        Parameters
        -----------
        port : string
            Serial port
        baudrate : int, default=19200
            baud rate for serial port
        """
        self._unit_equivalencies = unit_equivalencies
        self._emulated = emulate
        self.lock = threading.Lock()
        self.ser = None
        self.port = port
        self.baudrate = baudrate
        self._execute_immediately = True  # either collect commands into a list, or send one-by-one
        self._command_list = []  # list of commands to send together
        self.axis = [None, None, None]

    def connect(self):
        if self._emulated:
            self.ser = ESP301SerialEmulator(port=self.port, baudrate=self.baudrate,
                                            bytesize=8, timeout=1, parity='N', rtscts=1)
        else:
            self.ser = serial.Serial(port=self.port, baudrate=self.baudrate,
                                     bytesize=8, timeout=1, parity='N', rtscts=1)
        for i in range(3):
            try:
                self.axis[i] = NewportESP301Axis(
                    self, i+1, unit_equivalencies=self._unit_equivalencies)
            except (NewportError, serial.SerialException):
                # this will fail if axis is not enabled in controller
                self.axis[i] = None

    def disconnect(self):
        if self.ser is not None:
            self.ser.close()
        self.ser = None

    def __del__(self):
        self.disconnect()

    def _execute(self, cmd, errcheck):
        """
        Write string to device
        """
        response = None

        with self.lock:
            with NoKeyboardInterrupt():
                if not cmd.endswith('\r'):
                    cmd += '\r'
                self.ser.write(cmd.encode())
                if '?' in cmd:
                    response = self.ser.readline().strip()

                if errcheck:
                    self.ser.write(b'TB?\r')
                    error_string = self.ser.readline().decode().strip()
                    if not error_string.startswith('0'):
                        raise NewportError(error_string)

        return response.decode() if response is not None else ''

    def _cmd(self, cmd, params=tuple(), target=None, errcheck=True):
        """
        Here we wrap low-level read/write commands to allow specifying different axis and error checking

        Write to device.

        Parameters
        -----------
        cmd : string
            commands to send to device.
        params : iterable
            parameters for command, eg. position for move
        target : int or NewportESP301Axis, default None
            The destination axis. If none, msg is sent to controller
        errcheck : bool
            If true, the device is checked for errors after command is sent
            Note that error-checking is disabled in program mode, so ```errcheck```
            must be `False` in program mode
        """

        if isinstance(target, NewportESP301Axis):
            target = target.axis_idx

        raw_cmd = "{target}{cmd}{params}".format(
            target=target if target is not None else "",
            cmd=cmd.upper(),
            params=",".join(map(str, params))
        )
        response = None
        if self._execute_immediately:
            response = self._execute(raw_cmd, errcheck)
        else:
            self._command_list.append(raw_cmd)

        return response

    def reset(self):
        """
        Perform a hardware reset.

        Only effective if the watchdog timier is enabled by physical jumpers. See the users guide
        for more information
        """
        self._cmd("RS", errcheck=False)

    def abort_all_axes(self):
        self._cmd('AB')

    def stop_all_axes(self):
        """
        Stops motion on all axes, with programmed decceleration rate
        """
        self._cmd('ST')

    @property
    def version(self):
        return self._cmd('VE?')

    @contextmanager
    def define_program(self, program_id):
        """
        Erases existing program with given ID and records commands within this context to new program

        Parameters
        ----------
        program_id : int
            Label for new program

        Usage
        -----
        >>> controller = NewportESP301('/dev/ttyS0')
        >>> with controller.define_program(10):
        ...        controller.axis[0].move(0.001, absolute=False)
        ...        controller.axis[1].move(20, absolute=True)
        ...
        >>> controller.run_program(10)
        """
        if program_id not in range(1, 101):
            raise ValueError('program ID must be in range 1 to 100 (inclusive')

        # erase old program and enter programming mode
        self._cmd('XX', target=program_id)
        try:
            self._cmd('EP', target=program_id)
            yield
        finally:
            self._cmd('QP')

    @contextmanager
    def execute_bulk_commands(self, errcheck=True):
        """
        Context manager to send mutliple commands in a single message to controller.

        Note that it is not currently possible to receive the controllers response.

        Parameters
        -----------
        errcheck : bool
            Check for errors after each command

        Usage
        -------
        >>> controller = NewportESP301('/dev/ttyS0')
        >>> with controller.execute_bulk_command():
        ...     controller.axis[0].move(0.001, absolute=False)
        ...     controller.axis[1].move(20, absolute=True)
        ...
        """
        self._execute_immediately = False
        yield
        command_string = ';'.join(self._command_list)
        self._bulk_query_resp = self.execute(command_string, errcheck)
        self._command_list = []
        self._execute_immediately = True

    def run_program(self, program_id):
        """
        Run a previously defined program

        Program must be defined using ```NewportESP301.define_program```.

        Parameters
        ----------
        program_id : int
            ID number of previously defined program
        """
        if program_id not in range(1, 101):
            raise ValueError('program ID must be in range 1 to 100 (inclusive')

        self._cmd("EX", target=program_id)


class NewportESP301Axis:
    """
    Communication class for access to a single axis of an ESP301 controller.
    """

    # This dictionary encodes between the device unit code and the astropy unit
    # The device can operate in many units, and the unit can be set.
    # A unit equivalency is supplied with each instance to provide the
    # conversion between microstep and another physical unit
    _unit_dict = {
        0: u.def_unit('encoder count', u.dimensionless_unscaled),
        1: microstep,
        2: u.mm,
        3: u.micrometer,
        4: u.imperial.inch,
        5: u.def_unit('milliinch', u.imperial.inch / 1000),
        6: u.def_unit('microinch', u.imperial.inch / 1000000),
        7: u.deg,
        8: u.def_unit('gradian', 9*u.deg/10),
        9: u.rad,
        10: u.milliradian,
        11: u.microradian
    }

    acceleration = DeviceProperty('AC', units=1/u.s**2)
    deceleration = DeviceProperty('AG', units=1/u.s**2)
    estop_deceleration = DeviceProperty('AE', units=1/u.s**2)
    max_acceleration = DeviceProperty('AU', units=1/u.s**2)
    max_deceleration = DeviceProperty('AU', units=1/u.s**2)
    jerk = DeviceProperty('JK', units=1/u.s**3)
    velocity = DeviceProperty('VA', units=1/u.s)
    max_velocity = DeviceProperty('VU', units=1/u.s)
    max_base_velocity = DeviceProperty('VB', units=1/u.s)  # for stepper motors only
    jog_high_velocity = DeviceProperty('JH', units=1/u.s)
    jog_low_velocity = DeviceProperty('JL', units=1/u.s)
    homing_velocity = DeviceProperty('OH', units=1/u.s)
    position = DeviceProperty('TP', readonly=True)
    desired_position = DeviceProperty('DP', readonly=True)
    desired_velocity = DeviceProperty('DV', units=1/u.s)
    home_position = DeviceProperty('DH')
    left_limit = DeviceProperty('SL')
    right_limit = DeviceProperty('SR')
    error_threshold = DeviceProperty('FE')
    current = DeviceProperty('QI', units=u.A, relative_units=False)
    voltage = DeviceProperty('QV', units=u.V, relative_units=False)
    motor_type = IntegerDeviceProperty('QM')
    position_display_resolution = IntegerDeviceProperty('FP')  # 0--7, num. dec. places except 7=scientific notation
    # 1: trapezoidal, 2: s-curve, 3: jog, 4: slave to master's desired position
    # 5: slave to masters actual position, 6: slave to masters velocity
    trajectory_mode = IntegerDeviceProperty('TJ')
    microstep_factor = IntegerDeviceProperty('QS')
    accel_feedforward_gain = DeviceProperty('AF', units=u.dimensionless_unscaled, relative_units=False)
    proportional_gain = DeviceProperty('KP', units=u.dimensionless_unscaled, relative_units=False)
    derivative_gain = DeviceProperty('KD', units=u.dimensionless_unscaled, relative_units=False)
    integral_gain = DeviceProperty('KI', units=u.dimensionless_unscaled, relative_units=False)
    integral_gain_saturation = DeviceProperty('KS',
                                              units=u.dimensionless_unscaled,
                                              relative_units=False)

    def __init__(self, controller, axis_idx, unit_equivalencies=None):
        if not isinstance(controller, NewportESP301):
            raise ValueError('controller must be an instance of NewportESP301')

        self.controller = controller
        self.axis_idx = axis_idx
        # make a copy of the controller command function, with this axis index hard-wired
        self._cmd = partial(self.controller._cmd, target=self.axis_idx)
        self._unit_equivalencies = unit_equivalencies
        self._units = NewportESP301Axis._unit_dict[self._get_units()]

    def _format_get_command(self, code):
        return code + '?'

    def _format_set_command(self, code):
        return code

    @property
    def motion_complete(self):
        """
        True if all motion commands complete.
        """
        return bool(int(self._cmd("MD?")))

    # UNIT control. First private functions for talking to axis
    def _get_units(self):
        return int(self._cmd('SN?'))

    def _set_units(self, new_units):
        return self._cmd('SN', params=[int(new_units)])

    # and a utility function to lookup unit dictionary given an astropy unit
    def _get_unit_num(self, quantity):
        """
        Gets the newport integer label for a given unit
        """
        for num, quant in self._unit_dict.items():
            if quant == quantity:
                return num

        raise KeyError('{} is not a valid unit for NewportESP301Axis'.format(quantity))

    # now public property for axis through astropy units
    @property
    def units(self):
        """
        Units of axis as an ```~astropy.units.Quantity``` object.
        """
        return self._units

    @units.setter
    def units(self, newval):
        raise AttributeError('for unknown reasons, setting units does not work')
        if isinstance(newval, int):
            self._units = NewportError._unit_dict[newval]
        elif isinstance(newval, u.Quantity):
            self._units = newval
            newval = self._get_unit_num(newval)
            self._set_units(newval)
        self._set_units(newval)

    def home(self, search_mode=1, errcheck=True):
        """
        Move device to home.

        Parameters
        ----------
        search_mode : int
            0 : search by zero position count
            1 : search for home and index signals (default)
            2 : search only for home signal
            3 : search for positive limit signal
            4 : search for negative limit signal
            5 : seach for positive limit and index signals
            6 : search for negative limit and index signals

       Returns
        -------
        success: bool
            True or False depending on whether home command was successful        
        """
        try:
            self._cmd('OR', params=[search_mode], errcheck=errcheck)
        except Exception:
            return False
        return True

    def move(self, position, absolute=True, wait=False, block=False):
        """
        Move to or by a specified position.

        Parameters
        ----------
        position : float or ```~astropy.units.Quantity```
            The position to move to (if absolute) or amount to move by
            if relative. If a float is provided, we assume the position
            is given in the current controller units.
        absolute : bool
            If `False`, position is a relative offset from current position
        wait : bool
            If `True`, function returns immediately, but axis is instructed
            not to execute other commands until move is finished.
        block : bool
            IF `True`, function will not return until move is complete.

        Returns
        -------
        success: bool
            True or False depending on whether move command was successful
        """
        if not isinstance(position, u.Quantity):
            position = position * self.units
        position = position.to(self.units,
                               equivalencies=self._unit_equivalencies).value

        cmd = 'PA' if absolute else 'PR'
        try:
            self._cmd(cmd, params=[position])
            if wait:
                self.wait_for_position(position)
            if block:
                time.sleep(0.01)
                done = self.motion_complete
                while not done:
                    done = self.motion_complete
                    time.sleep(0.01)
        except Exception:
            return False
        return True

    def wait_for_position(self, position):
        """
        Wait for axis to reach given position before executing other commands.

        Following this call, other commands can be sent to the controller, but they
        will not be executed until the specified position is reached. Most useful
        in bulk command mode or programming mode.

        Parameters
        ----------
        position : float or ```~astropy.units.Quantity```
            The position to wait for. If a float is provided, we assume the position
            is given in the current controller units.
        """
        if not isinstance(position, u.Quantity):
            position = position * self.units
        position = position.to(self.units, equivalencies=self._unit_equivalencies).value
        self._cmd('WP', params=[position])

    def wait_for_stop(self):
        """
        Waits for axis to stop before executing other commands.

        Following this call, other commands can be sent to the controller, but they
        will not be executed until the axis has stopped. Most useful
        in bulk command mode or programming mode.
        """
        self._cmd('WS')

    def abort(self):
        """
        Stops motion immediately
        """
        self._cmd('AB')

    def stop(self):
        """
        Stops motion with programmed decceleration.

        Returns
        -------
        success: bool
            True or False depending on whether stop command was successful
        """
        try:
            self._cmd('ST')
        except Exception:
            return False
        return True

    def move_to_hardware_limit(self):
        """
        Move to hardware travel limit
        """
        self._cmd('MT')

    def move_indefinitely(self):
        """
        Move until a stop request is received
        """
        self._cmd('MV')

    def power_on(self):
        """
        Turns on motor axis
        """
        self._cmd("MO")

    def power_off(self):
        """
        Turns off motor axis
        """
        self._cmd("MF")

    def poll(self):
        """
        State machines will want to automatically poll position.

        Update current position and return true/false for moving state
        """
        moving = not self.motion_complete
        return self.position, moving

    def on_target(self, rtol=1.0e-7):
        """
        Have we reached target sucessfully?

        Also polled by state machines

        TODO: make this work with conversion to microstep
        """
        current = self.position
        target = self.desired_position
        delta = (current-target).to_value(current.unit)
        delta = abs((delta/current).decompose().value)

        return self.motion_complete and delta < rtol


class ESP301SerialEmulator(FakeSerial):
    simple_commands = {
        'TB?': '0',
        'VE?': 'simulator',
        'MO': '',
        'MF': ''
    }

    def __init__(self, *args, **kwargs):
        super(ESP301SerialEmulator, self).__init__(*args, **kwargs)
        axis1 = AxisEmulator(2000)  #  2000 microstep/sec
        axis2 = AxisEmulator(2000)
        axis3 = AxisEmulator(2000)
        self.axes = [axis1, axis2, axis3]
        self.axis_unit_codes = [7, 7, 2]
        self.axis_units = [u.deg, u.deg, u.mm]
        self.equiv = [
            # from, to, forward, backward
            (u.deg, microstep, lambda x: x*1000, lambda x: x/1000),
            (u.mm, microstep, lambda x: x*1000, lambda x: x/1000),
        ]

    def _convert_to_microstep(self, value):
        if not isinstance(value, u.Quantity):
            # assume native unit for axis
            value = value * self.axis_units[self._axis-1]

        return value.to_value(microstep, equivalencies=self.equiv)

    def _convert_from_microstep(self, value, unit):
        if not isinstance(value, u.Quantity):
            # assume microstep
            value = value * microstep
        return value.to_value(unit, equivalencies=self.equiv)

    @FakeSerial.register_command('SN?', None)
    def _get_unit(self):
        return self.axis_unit_codes[self._axis-1]

    @FakeSerial.register_command('PA', float)
    def _move_axis(self, position):
        axis = self.axes[self._axis-1]
        ms = self._convert_to_microstep(position)
        axis.target_position = int(ms)

    @FakeSerial.register_command('OR', None)
    def _home_axis(self, args):
        self._move_axis(0)

    @FakeSerial.register_command('MD?', None)
    def _is_moving(self):
        axis = self.axes[self._axis-1]
        return int(not axis.moving)

    @FakeSerial.register_command('ST', float)
    def _stop_axis(self):
        if self._axis is None:
            # stop all
            for axis in self.axes:
                axis.moving = False
        else:
            axis = self.axes[self._axis-1]
            axis.moving = False

    @FakeSerial.register_command('TP?', None)
    def _axis_pos(self):
        axis = self.axes[self._axis-1]
        unit = self.axis_units[self._axis-1]
        ms = axis.current_position
        return self._convert_from_microstep(ms, unit)

    @FakeSerial.register_command('DP?', None)
    def _axis_targpos(self):
        axis = self.axes[self._axis-1]
        unit = self.axis_units[self._axis-1]
        ms = axis.target_position
        return self._convert_from_microstep(ms, unit)

    def _process_received(self):
        self._receivedData = self._receivedData.strip('\r')
        try:
            self._axis = int(self._receivedData[0])
            self._receivedData = self._receivedData[1:]
        except ValueError:
            self._axis = None
        super(ESP301SerialEmulator, self)._process_received()



from hcam_devices.machines import get_yaml_file
from sismic.io import export_to_plantuml, import_from_yaml
from plantuml import PlantUML

import sys

name = sys.argv[1]
filename = get_yaml_file(name)
print(filename)
machine = import_from_yaml(filepath=filename)
plantuml = export_to_plantuml(machine)

maker = PlantUML(url='http://www.plantuml.com/plantuml/img/')
png_data = maker.processes(plantuml)

with open(name.replace('yaml', 'png'), 'wb') as out:
    out.write(png_data)



# COMPO components
import pickle
import astropy.units as u
from astropy.time import Time
import traceback
from twisted.logger import Logger

from ..devices.properties import microstep
from ..devices.newport import NewportESP301
from .slide import LinearStage
from ..machines import create_machine


MIN_MS = 0
MAX_MS = 1049869  # 25 mm travel
MM_PER_MS = 2.38125e-05
MIN_MM = MM_PER_MS * MIN_MS
MAX_MM = MM_PER_MS * MAX_MS


def mm_to_ms(value):
    return MIN_MS + int((MAX_MS - MIN_MS) * (value - MIN_MM) / (MAX_MM - MIN_MM) + 0.5)


def ms_to_mm(value):
    return MIN_MM + (MAX_MM - MIN_MM) * (value - MIN_MS) / (MAX_MS - MIN_MS)


equivalencies = [
    # from, to, forward, backward (assume 1-1 for now)
    (u.rad, microstep, lambda x: x, lambda x: x),
    (u.mm, microstep, mm_to_ms, ms_to_mm),
]

# match order in NewportESP301
esp_axes = {1: "pickoff", 2: "injection"}


class Compo(object):
    """
    Model for COMPO.

    COMPO contains a linear stage and two rotational arms. Each of
    these have their own hardware model. This is a convenience model
    that listens to, and forwards, telemetry from these bits of kit.

    The correct topics to subscribe to are passed as parameters.
    """

    log = Logger()

    def __init__(self, arm_channel, lens_channel):
        # we subscribe to telemetry on these channels
        self.subscribed_channels = {"lens": lens_channel, "arms": arm_channel}
        # pass through topics merely pass any published messages
        # on to a different topic
        self.pass_through_topics = {
            "target_injection_angle": f"{arm_channel}.target_injection_angle",
            "target_pickoff_angle": f"{arm_channel}.target_pickoff_angle",
            "target_lens_position": f"{lens_channel}.target_position",
        }

        self._telemetry_timestamps = {"lens": Time.now(), "arms": Time.now()}

        # this is the basic package we publish as telemetry
        self._telemetry_pkg = {
            "state": {"arms_state": "", "lens_state": ""},
            "pickoff_angle": {"target": 0, "current": 0},
            "injection_angle": {"target": 0, "current": 0},
            "lens_position": {"target": 0, "current": 0},
        }

    def on_received_telemetry(self, data):
        """
        Handle telemetry data. Could come from any
        subscribed topics
        """
        telemetry = pickle.loads(data)
        telemetry_time = telemetry["timestamp"]
        device_status = telemetry["state"]["connection"]

        if "pickoff_angle" in telemetry:
            source = "arms"
        elif "position":
            source = "lens"
        else:
            raise ValueError("unexpected telemetry pkg: ", telemetry)
        self._telemetry_timestamps[source] = telemetry_time

        if source == "arms":
            # set telemetry package for arms
            self._telemetry_pkg["state"]["arms_state"] = telemetry["state"]
            self._telemetry_pkg["pickoff_angle"] = telemetry["pickoff_angle"]
            self._telemetry_pkg["injection_angle"] = telemetry["injection_angle"]
        else:
            self._telemetry_pkg["state"]["lens_state"] = telemetry["state"]
            self._telemetry_pkg["lens_position"] = telemetry["position"]

    def telemetry(self):
        # use stalest timestamp we have as the data timestamp
        oldest_ts = min(self._telemetry_timestamps, key=self._telemetry_timestamps.get)
        tel = dict(timestamp=self._telemetry_timestamps[oldest_ts])
        tel.update(self._telemetry_pkg)
        return tel


class CompoLens(LinearStage):
    """
    COMPO contins a Zaber linear stage which controls the lens.
    """

    def __init__(self, ip_address, port, emulate=True):
        super().__init__(ip_address, port, equivalencies, emulate)
        self.natural_units = u.mm

        # override the maximum position to avoid querying it, which is unreliable
        self.dev.max_position = 1049869


class CompoArms(object):
    """
    Model for COMPO rotational arms.

    COMPO is controlled by a Newport ESP301 controller, which has 3 axes.
    The first 2 axes are the angles of the pickoff arm and injection arm.
    """

    log = Logger()

    # can be set by publishing to a device level topic of same name
    settable_attributes = ("target_injection_angle", "target_pickoff_angle")

    # all state change triggers are automatically addded as RPC calls
    # add any extra methods you want exposed here
    extra_rpc_calls = ("power_on_axis", "power_off_axis")

    def __init__(self, esp_port, emulate=False):
        self.esp = NewportESP301(
            esp_port, emulate=emulate, unit_equivalencies=equivalencies
        )

        # state machines to track state. Controller for connections
        self._controller = create_machine(
            "connections", initial_context={"device": self, "name": "compo"}
        )
        self.machines = dict(connection=self._controller)

        # newport axes!
        # ORDER HERE MUST MATCH ORDER IN Newport!
        for ax in esp_axes:
            name = esp_axes[ax]
            ctx = {
                "device": self.esp.axis[ax - 1],
                "current": 0 * u.mm if name == "lens" else 0 * u.deg,
                "target": 0 * u.mm if name == "lens" else 0 * u.deg,
                "poll_time": 2,
            }
            axis = create_machine("motors", initial_context=ctx)
            self.machines[name] = axis
            # bind axes to receive connect disconnect from controller
            self._controller.bind(axis)

    @property
    def current_injection_angle(self):
        return self.machines["injection"].context["current"]

    @property
    def target_injection_angle(self):
        return self.machines["injection"].context["target"]

    @target_injection_angle.setter
    def target_injection_angle(self, value):
        print("setting injection angle")
        if not isinstance(value, u.Quantity):
            value = value * u.deg
        self.machines["injection"].queue("positionSet", position=value)

    @property
    def current_pickoff_angle(self):
        return self.machines["pickoff"].context["current"]

    @property
    def target_pickoff_angle(self):
        return self.machines["pickoff"].context["target"]

    @target_pickoff_angle.setter
    def target_pickoff_angle(self, value):
        if not isinstance(value, u.Quantity):
            value = value * u.deg
        self.machines["pickoff"].queue("positionSet", position=value)

    def telemetry(self):
        """
        Called periodically to provide a telemetry package for the device
        """
        ts = Time.now()
        tel = dict(timestamp=ts)
        state = {key: self.machines[key].configuration for key in self.machines}
        tel["state"] = state
        tel["pickoff_angle"] = dict(
            current=self.current_pickoff_angle, target=self.target_pickoff_angle
        )
        tel["injection_angle"] = dict(
            current=self.current_injection_angle, target=self.target_injection_angle
        )
        return tel

    # Methods REQUIRED for controller state machine
    def try_connect(self):
        try:
            # connect to newport
            self.esp.connect()
            # update device in state machines for arms
            for ax in esp_axes:
                name = esp_axes[ax]
                machine = self.machines[name]
                axis = self.esp.axis[ax - 1]
                machine.context["device"] = axis
        except Exception as err:
            print(traceback.format_exc())
            return False
        return True

    def disconnect(self):
        self.esp.disconnect()

    # additional RPC calls
    def power_on_axis(self, axis):
        self.esp.axis[axis].power_on()

    def power_off_axis(self, axis):
        self.esp.axis[axis].power_off()



import CosNaming
from omniORB import CORBA

# not part of this module, copy modules from GCS and place in PYTHONPATH
import HIPERTELESCOPESERVER
import os

from ..wamp.utils import call


class TelescopeServer(object):
    def __init__(self):
        self.establish_connection()

    def establish_connection(self):
        conn = CORBAConnection()

        conn.init_orb()
        conn.resolve_nameservice()

        # name of CORBA object to access telescope server
        name = [CosNaming.NameComponent(id="HIPERTelescopeServer", kind="")]

        self._server = conn.get_object(
            name, HIPERTELESCOPESERVER.HIPERTelescopeServer_ifce
        )

    @property
    def ready(self):
        try:
            self._server.getTelescopeParams()
        except:
            return False
        return True

    def in_position(self):
        """
        Is telescope in position and ready to observe?
        """
        try:
            ready = self._server.areMACSFollowErrorBelowThreshold()
            # has mirror settled?
            ready = ready and self._server.isM1NoiseBelowThreshold()
        except:
            # failed to read status, don't assume OK.
            ready = False

        return ready

    def requestTelescopeOffset(self, dra, ddec):
        """
        Request a telescope offset
        """
        self._server.requestTelescopeOffset(dra, ddec)

    # pass all unknown method lookups to server
    def __getattr__(self, name):
        return getattr(self._server, name)


class CORBAConnection(object):
    """
    A class to handle communication with the GTC CORBA system.

    Simply delegates the relevant operations to the current state object.
    """

    def __init__(self):
        self.new_state(RawCORBAState)

    def new_state(self, state):
        # change the class to a concrete type
        self._state = state

    def init_orb(self):
        self._state.init_orb(self)

    def resolve_nameservice(self):
        self._state.resolve_nameservice(self)

    def get_object(self, binding_name, cls):
        """
        Get a reference to a remote object using CORBA
        """
        return self._state.get_object(self, binding_name, cls)


class RawCORBAState(object):
    """
    A CORBA state just after creation, with no ORB or nameservice
    """

    @staticmethod
    def init_orb(conn):
        # setup CORBA, TODO put correct GTC host here
        os.environ["ORBDefaultInitRef"] = "corbaname::161.72.88.75:12008"
        os.environ["ORBclientCallTimeOutPeriod"] = "2000"  # in milliseconds
        orb = CORBA.ORB_init([], CORBA.ORB_ID)
        conn.orb = orb
        conn.new_state(InitialisedCORBAState)

    @staticmethod
    def resolve_nameservice(conn):
        raise RuntimeError("ORB not initialised")

    @staticmethod
    def get_object(conn, binding_name, cls):
        """
        Get a reference to a remote object using CORBA
        """
        raise RuntimeError("ORB not initialised")


class InitialisedCORBAState(object):
    """
    A CORBA state just after creation, with no ORB or nameservice
    """

    @staticmethod
    def init_orb(conn):
        raise RuntimeError("ORB already initialised")

    @staticmethod
    def resolve_nameservice(conn):
        # get instance of nameserver from GTC CORBA sever
        try:
            obj = conn.orb.resolve_initial_references("NameService")
            rootContext = obj._narrow(CosNaming.NamingContext)
        except CORBA.TRANSIENT:
            raise IOError("could not connect to GTC CORBA NameServer")

        if rootContext is None:
            raise IOError("attempt to get NameServer from GTC failed")
        conn.rootContext = rootContext
        conn.new_state(ReadyCORBAState)

    @staticmethod
    def get_object(conn, binding_name, cls):
        """
        Get a reference to a remote object using CORBA
        """
        raise RuntimeError("NameService needs to be resolved first")


class ReadyCORBAState(object):
    """
    A CORBA state which is ready to retrieve objects
    """

    @staticmethod
    def init_orb(conn):
        raise RuntimeError("ORB already initialised")

    @staticmethod
    def resolve_nameservice(conn):
        # get instance of nameserver from GTC CORBA sever
        raise RuntimeError("NameService already resolved")

    @staticmethod
    def get_object(conn, binding_name, object_cls):
        """
        Get a reference to a remote object using CORBA
        """
        try:
            obj = conn.rootContext.resolve(binding_name)
            narrowed = obj._narrow(object_cls)
        except CORBA.TRANSIENT:
            raise IOError("Attempt to retrieve object failed")

        if narrowed is None:
            raise IOError("Attempt to retrieve object got a different class of object")

        return narrowed



from pathlib import Path

from sismic.interpreter import Interpreter
from sismic.clock import UtcClock
from sismic.io import import_from_yaml

try:
    from importlib import resources as importlib_resources
except Exception:
    # backport for python 3.6
    import importlib_resources


def get_yaml_file(machine_name):
    if not machine_name.endswith(".yaml"):
        machine_name += ".yaml"

    try:
        full_path = (
            importlib_resources.files("hcam_devices") / "machines" / machine_name
        )
        if not full_path.exists():
            # yaml file not installed with package, see if it is specified directly
            full_path = Path(machine_name)
    except AttributeError:
        # hack for Python <3.9
        from pkg_resources import resource_filename, resource_exists

        full_path = resource_filename("hcam_devices", f"machines/{machine_name}")
        if not resource_exists("hcam_devices", f"machines/{machine_name}"):
            # yaml file not installed with package, see if it is specified directly
            full_path = Path(machine_name)

    return str(full_path)


def create_machine(name, **kwargs):
    full_path = get_yaml_file(name)
    return Interpreter(import_from_yaml(filepath=full_path), clock=UtcClock(), **kwargs)



import os
import time
import six

from astropy.io import ascii
from astropy.io import fits
from astropy.time import Time
from twisted.logger import Logger

from ..utils.obsmodes import Idle, get_obsmode
from ..devices.ngc import (NGCDevice, NGCEmulator)
from ..machines import create_machine


class NGC(object):
    """
    Model for an ESO NGC controller
    """
    log = Logger()

    settable_attributes = ()
    extra_rpc_calls = ('start', 'stop', 'abort', 'pon', 'poff',
                       'seq_start', 'seq_stop', 'trigger', 'reset',
                       'summary', 'status', 'setup',
                       'online', 'standby', 'offline', 'exit',
                       'start_ngc_server', 'stop_ngc_server', 'setup_ngc_server',
                       'load_setup', 'add_hdu', 'cmd')

    def __init__(self, emulate=True):
        if emulate:
            self._dev = NGCEmulator()
        else:
            self._dev = NGCDevice()

        # NGC already has it's own state machine, so don't duplicate here
        self.machines = {}

    def online(self):
        """
        Moves NGC server into ONLINE state.

        When moving to this state the h/w is connected and the controller reset.
        """
        return self.cmd('online')

    def exit(self):
        """
        Shuts down NGC server.

        Once in this state it is impossible to talk to NGC server and it must be started
        using `NGC.start_ngc_server`
        """
        return self.cmd('exit')

    def offline(self):
        """
        Turns NGC server off, moving it into LOADED state.

        In this state, the server is running and you can communicate with it, but all sub-processes
        (e.g acquisition task) are shut down and hardware is disconnected.
        """
        return self.cmd('off')

    def standby(self):
        """
        Move NGC server to STANDBY state.

        In this state, the server is not connected to the hardware, but the sub-processes 
        (e.g acquisition task) are enabled.
        """
        return self.cmd('standby')

    def stop_ngc_server(self):
        """
        Shuts down ESO NGC Server
        """
        try:
            output = self._dev.stop_ngc_sw()
        except Exception as e:
            self.log.error('could not start NGC software: ' + str(e))
            return False

        for key in output:
            msg = output[key]
            self.log.info("cmd {}: {}".format(key, msg))
        return True

    def start_ngc_server(self):
        """
        Run commands necessary to bring up the ESO NGC server
        """
        try:
            output = self._dev.start_ngc_sw()
        except Exception as e:
            self.log.error('could not start NGC software: ' + str(e))
            return False

        for key in output:
            msg = output[key]
            self.log.info("cmd {}: {}".format(key, msg))
        return True

    def setup_ngc_server(self):
        """
        Load HiPERCAM specific settings
        """
        try:
            output = self._dev.setup_ngc_hipercam()
        except Exception as e:
            self.log.error('could not setup NGC for hipercam: ' + str(e))
            return False

        for key in output:
            msg = output[key]
            self.log.info("cmd {}: {}".format(key, msg))
        return True

    def cmd(self, cmd, *args):
        self.log.info("cmd {}: {}".format(cmd, ' '.join(args)))
        try:
            msg, cmd_ok = self._dev.send_command(cmd, *args)
            if not cmd_ok:
                self.log.error('cmd {} failed with msg {}'.format(cmd, msg))
        except Exception as e:
            msg = 'could not run command {}: {}'.format(cmd, e)
            self.log.error('could not run command {}: {}'.format(cmd, e))
            cmd_ok = False

        return msg, cmd_ok

    def telemetry(self):
        """
        Called periodically to provide status etc for the NGC
        """
        ts = Time.now()
        tel = dict(timestamp=ts)
        tel.update(self._dev.database_summary)
        frame_msg, ok = self._dev.send_command('status', 'DET.FRAM2.NO')
        try:
            frame_no = int(frame_msg.split()[1])
        except:
            frame_no = 0
        if not ok:
            tel['exposure.frame'] = 0
        else:
            tel['exposure.frame'] = frame_no
        return tel

    def start(self, *args):
        return self.cmd('start')

    def stop(self, *args):
        return self.cmd('stop')

    def abort(self, *args):
        return self.cmd('abort')

    def pon(self, *args):
        return self.cmd('cldc 0 enable')

    def poff(self, *args):
        return self.cmd('cldc 0 disable')

    def seq_start(self, *args):
        return self.cmd('seq 0 start')

    def seq_stop(self, *args):
        return self.cmd('seq 0 stop')

    def trigger(self, *args):
        return self.cmd('seq trigger')

    def reset(self, *args):
        return self.cmd('reset')

    def summary(self, *args):
        return self._dev.database_summary

    def status(self, param_name=None):
        if param_name is None:
            return self.cmd('status')

        response, response_ok = self.cmd('status', param_name)
        if not response_ok:
            return response, response_ok

        try:
            name, val = response.split()
            if name != param_name:
                msg = 'unexpected response: {} vs {}'.format(
                    name, param_name
                )
                return msg, False

            return val, response_ok
        except Exception as err:
            return str(err), False

        return self.cmd('status', param_name)

    def setup(self, param_name, value):
        return self.cmd('setup', param_name, value)

    def load_setup(self, data):
        """
        Post a JSON setup to the server

        Parameters
        -----------
        data : dict
            setup info
        """
        try:
            obsmode = get_obsmode(data)
        except Exception as err:
            self.log.error('could not parse obsmode data: ' + str(err))
            return False, 'could not parse obsmode data'

        resp = self.cmd('seq stop')
        if not resp:
            self.log.error('could not stop sequencer')
            return False, 'could not stop sequencer'

        time.sleep(0.1)
        resp = self.cmd(obsmode.readmode_command)
        if not resp:
            self.log.error('could not set read mode: ' + resp)
            return False, 'could not set read mode: ' + resp

        time.sleep(0.1)
        resp = self.cmd(obsmode.acq_command)
        if not resp:
            self.log.error('could not start acquisition process: ' + resp)
            return False, 'could not start acquisition process: ' + resp

        time.sleep(0.1)
        for header_command in obsmode.header_commands:
            resp = self.cmd(header_command)
        if not resp:
            msg = 'could not set header param {}: {} '.format(
                header_command, resp
            )
            self.log.error(msg)
            return False, msg

        time.sleep(0.1)
        resp = self.cmd(obsmode.setup_command)
        if not resp:
            self.log.error('could not setup run: ' + resp)
            return False, 'could not setup run: ' + resp

        if isinstance(obsmode, Idle):
            """
            A run start will start the sequencer, saving data and clearing the chips.
            There is no run start when we switch to idle mode, so start the sequencer
            manually
            """
            resp = self.cmd('seq start')
            if not resp:
                self.log.error('could not start sequencer: ' + resp)
                return False, 'could not start sequencer: ' + resp

        return True, 'All OK'

    def add_hdu(self, table_data, run_number):
        """
        Handle an uploaded table, and append it to the appropriate FITS file

        Parameters
        -----------
        table_data: bytes
            a binary encoded VOTable, which we will write to a FITS HDU,
            which is appended onto the original FITS file.
        run_number: int
            the run to append the table data to
        """
        try:
            t = ascii.read(table_data, format='ecsv')
        except Exception as e:
            self.log.error('cannot decode table data: ' + str(e))
            return

        filename = os.path.join('/data', "run{:04d}.fits".format(run_number))
        if not os.path.exists(filename):
            self.log.error('no such filename exists: ' + filename)
            return

        try:
            new_hdu = fits.table_to_hdu(t)
            existing_hdu_list = fits.open(
                filename, mode='append', memmap=True
            )
            existing_hdu_list.append(new_hdu)
            existing_hdu_list.flush()
        except Exception as e:
            self.log.error('could not write HDU to run: ' + str(e))



# Utility to query and set temps via Ethernet on Meerstetter
from __future__ import absolute_import, unicode_literals, print_function, division
import socket
import random
import six
import struct
import threading
import logging

from ..testing.sensors import SensorEmulator

DEFAULT_TIMEOUT = 5
TEC_CURRENT_LIMIT = 10.7


error_codes = {
    "+01": "Command not available",
    "+02": "Device is busy",
    "+03": "Communication error",
    "+04": "Format error",
    "+05": "Parameter not available",
    "+06": "Parameter is read only",
    "+07": "Value is out of range",
    "+08": "Instance not available",
}

STATUS_CODES = {
    0: "init",
    1: "ready",
    2: "run",
    3: "error",
    4: "bootloader",
    5: "resetting",
}


def hex_to_int(hexstring):
    return int(hexstring, 16)


def hex_to_float32(hexstring):
    try:
        if six.PY2:
            byteval = hexstring.strip().decode("hex")
        else:
            byteval = bytes.fromhex(hexstring.strip())
    except Exception as err:
        raise RuntimeError("cannot decode string {}!\n{}".format(hexstring, err))
    return struct.unpack(str("!f"), byteval)[0]


def float32_to_hex(f):
    return format(struct.unpack(str("<I"), struct.pack(str("<f"), f))[0], "X")


class CRCCalculator(object):
    def __init__(self):
        self.polynomial = 0x1021
        self.preset = 0
        self._lut = [self._calc_initial(i) for i in range(256)]

    def _calc_initial(self, c):
        crc = 0
        c = c << 8
        for j in range(8):
            if (crc ^ c) & 0x8000:
                crc = (crc << 1) ^ self.polynomial
            else:
                crc = crc << 1
            c = c << 1
        return crc

    def _update_crc(self, crc, c):
        cc = 0xFF & c

        tmp = (crc >> 8) ^ cc
        crc = (crc << 8) ^ self._lut[tmp & 0xFF]
        crc = crc & 0xFFFF
        return crc

    def __call__(self, msg):
        crc = self.preset
        for c in msg:
            crc = self._update_crc(crc, ord(c))
        return format(crc, "0>4X")


class MeerstetterTEC1090Emulator(object):
    def __init__(self, address, port):
        self._heatsinks = {i: SensorEmulator(38, 0.1, 0.001) for i in range(1, 4)}
        self._currents = {i: SensorEmulator(8, 0.1, 0.001) for i in range(1, 4)}
        self._temperatures = {i: SensorEmulator(-90, 0.1, 0.01) for i in range(1, 4)}

    def disconnect(self):
        return True

    def connect(self):
        return True

    def get_status(self, addr):
        return 2

    def get_heatsink_temp(self, addr):
        return self._heatsinks[addr].current_value

    def set_ccd_temp(self, addr, val):
        self._temperatures[addr].target_value = val

    def get_ccd_temp(self, addr):
        return self._temperatures[addr].current_value

    def get_setpoint(self, addr):
        return self._temperatures[addr].target_value

    def get_current(self, addr):
        return self._currents[addr].current_value


class MeerstetterTEC1090(object):
    """
    Class to use TCP/IP to communicate with TEC-1090 controllers
    mounted in LTR-1200. Communication protocol is MeCom.

    See MeCom protocol spec document 5117B and TEC protocol document
    5136 for details.
    """

    def __init__(self, address, port, file_logging=False):
        self.logger = logging.getLogger("meerstetter.{}".format(address))
        self.logger.setLevel(logging.DEBUG)
        if file_logging:
            fh = logging.FileHandler("meerstetter.log")
            fh.setLevel(logging.DEBUG)
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(threadName)s - %(message)s"
            )
            fh.setFormatter(formatter)
            self.logger.addHandler(fh)

        self.address = address
        self.port = port
        self.seq_no = random.randint(1, 1000)
        self.crc_calc = CRCCalculator()
        self.tec_current_limit = TEC_CURRENT_LIMIT
        self._lock = threading.Lock()
        self.sock = None

    def __del__(self):
        self.disconnect()

    def disconnect(self):
        self.sock.close()
        self.sock = None

    def connect(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(DEFAULT_TIMEOUT)
            s.connect((self.address, self.port))
            welcome = s.recv(1024)
            if "Welcome" not in welcome.decode():
                self.logger.error("did not receive welcome message from meerstetter")
                raise IOError("did not receive welcome message from meerstetter")
        except Exception:
            # connection failed
            return False
        self.sock = s
        return True

    def _assemble_frame(self, address, payload):
        """
        Assemble the frame of data to send.

        Transmissions are sent in blocks of data known as frames. Each frame has
        several fields:

        1: Control (source); ASCII Char; 1 bytes
        2: Address; UINT8; 2 bytes
        3: Sequence No; UINT16; 4 bytes
        4: Payload; N bytes
        5: Checksum, 4 bytes
        6: End of Frame, 1 byte <CR>
        """
        cs = "#"
        addr = format(address, "0>2X")
        seq = format(self.seq_no, "0>4X")
        # increment sequence number, but ensure it's not > 65535
        # that would overflow buffer! We roll over at 65534
        self.seq_no = (self.seq_no + 1) % 65535
        msg = cs + addr + seq + payload
        eof = "\r"
        return msg + self.crc_calc(msg) + eof

    def _send_frame(self, frame_msg):
        # threadsafe and error-safe opening of socket
        with self._lock:
            self.sock.send(frame_msg.encode())
            ret_msg = self.sock.recv(1024)
        ret_msg = ret_msg.decode().strip()
        self.logger.info("got ret_msg = {}".format(ret_msg))
        self._check_response(frame_msg, ret_msg)
        return self._strip_response(ret_msg)

    def _check_response(self, frame_msg, ret_msg):
        if ret_msg[0] == "!" and frame_msg[1:7] == ret_msg[1:7]:
            # a valid response, and same seq no. so far so good
            crc_back = ret_msg[-4:]
            crc_out = frame_msg[-5:-1]
            package_in = ret_msg[:-4]
            if self.crc_calc(package_in) != crc_back and crc_out != crc_back:
                raise IOError(
                    "checksum of return message not OK:\nOut: {}\nBack: {}".format(
                        frame_msg, ret_msg
                    )
                )
        else:
            raise IOError(
                "response not from device, or sequence number mismatch:\nOut: {}\nBack: {}".format(
                    frame_msg, ret_msg
                )
            )

    def _strip_response(self, ret_msg):
        return ret_msg[7:-4]

    def get_param(self, address, param_no, instance, param_type="float"):
        payload = "?VR{param_no:0>4X}{instance:0>2X}".format(
            param_no=param_no, instance=instance
        )
        self.logger.info("payload = {}".format(payload))
        frame_msg = self._assemble_frame(address, payload)
        encoded_param_val = self._send_frame(frame_msg)
        if encoded_param_val == "+05":
            raise IOError("param {} not available".format(param_no))

        if encoded_param_val in error_codes:
            raise IOError(
                "failed to get param {}\n{}".format(
                    param_no, error_codes[encoded_param_val]
                )
            )

        if param_type == "float":
            return hex_to_float32(encoded_param_val)
        else:
            return hex_to_int(encoded_param_val)

    def reset_tec(self, address):
        payload = "RS"
        frame_msg = self._assemble_frame(address, payload)
        ret_msg = self._send_frame(frame_msg)
        if ret_msg != "":
            raise IOError("unexpected response to reset command {}".format(ret_msg))

    def set_ccd_temp(self, address, target_temp):
        param_no = 3000
        payload = "VS{param_no:0>4X}{instance:0>2X}{encoded_val}".format(
            param_no=param_no, instance=1, encoded_val=float32_to_hex(target_temp)
        )
        frame_msg = self._assemble_frame(address, payload)
        ret_msg = self._send_frame(frame_msg)
        if ret_msg.startswith("+"):
            if ret_msg in error_codes:
                raise IOError(error_codes[ret_msg])
            else:
                raise IOError("unknown error code {}".format(ret_msg))

    def get_ccd_temp(self, address):
        self.logger.info("getting CCD{} temp".format(address))
        param_no = 1000
        return self.get_param(address, param_no, 1)

    def get_setpoint(self, address):
        param_no = 3000
        return self.get_param(address, param_no, 1)

    def get_heatsink_temp(self, address):
        param_no = 1001
        return self.get_param(address, param_no, 1)

    def get_power(self, address):
        current = self.get_param(address, 1020, 1)
        voltage = self.get_param(address, 1021, 1)
        return current * voltage

    def get_current(self, address):
        return self.get_param(address, 1020, 1)

    def get_status(self, address):
        param_no = 104
        status = self.get_param(address, param_no, 1, param_type="int")
        return status



import pickle

from astropy.time import Time
from astropy import units as u
from twisted.logger import Logger

STALE_TELEMETRY_THRESHOLD = 60


class CCDHead(object):
    """
    Model for a HiPERCAM CCD head.

    This is a convenience model that listens to, and forwards,
    telemetry from the vacuum gauge, flow sensor and peltier
    controller for the relevant CCD head.

    The correct topics to subscribe to are passed as parameters
    """
    log = Logger()

    def __init__(self, pressure_channel, flow_channel,
                 peltier_channel,
                 pen_number, meerstetter_channel):

        self.pen_number = str(pen_number)
        self.meerstetter_channel = str(meerstetter_channel)

        # we subscribe to telemetry on these channels
        self.subscribed_channels = {
            'pressure': pressure_channel,
            'flow': flow_channel,
            'peltier': peltier_channel
        }

        # pass through topics merely pass any published messages on to a different topic
        setpoint_topic = 'target_temperature' + self.meerstetter_channel
        self.pass_through_topics = {
            'setpoint': '.'.join([self.subscribed_channels['peltier'], setpoint_topic])
        }

        self._telemetry_timestamps = {
            'pressure': Time.now(),
            'flow': Time.now(),
            'peltier': Time.now()
        }

        # this is the basic package we publish as telemetry
        self._telemetry_pkg = {
            'pressure': 0.0,
            'flow': 0.0,
            'heatsink': 0.0,
            'peltier_current': 0.0,
            'peltier_status': 'error',
            'setpoint': 0.0,
            'temperature': 0.0,
            'state': "ERR",
            'state_reason': 'OK'
        }

    def _check_state(self):
        """
        check current values are OK
        """
        state = "OK", 'None'

        if self._telemetry_pkg['pressure'].to_value(u.mbar) > 8.0e-3:
            state = 'ERR', 'high pressure'
        flow = self._telemetry_pkg['flow']
        if flow < 0.4 * u.liter/u.minute:
            state = 'ERR', 'low flow'
        deltaT = abs(self._telemetry_pkg['temperature'] - self._telemetry_pkg['setpoint'])
        deltaT = deltaT.to_value(u.deg_C, equivalencies=u.temperature())
        if deltaT > 2:
            state = 'WARN', 'temperature not near setpoint'
        if self._telemetry_pkg['peltier_status'] != 'run':
            state = 'ERR', 'peltier not running'
        for source, telemetry_time in self._telemetry_timestamps.items():
            telemetry_age = (Time.now() - telemetry_time).to_value(u.s)
            if telemetry_age > STALE_TELEMETRY_THRESHOLD:
                state = 'ERR', 'stale telemetry from {}'.format(source)

        return state

    def on_received_telemetry(self, data):
        """
        Handle telemetry data. Could come from any
        subscribed topics
        """
        telemetry = pickle.loads(data)
        telemetry_time = telemetry['timestamp']
        device_status = telemetry['state']['connection']

        if 'pen1' in telemetry:
            source = 'flow'
        elif 'pressure' in telemetry:
            source = 'pressure'
        elif 'heatsink1' in telemetry:
            source = 'peltier'
        else:
            raise ValueError('unexpected telemetry pkg: ', telemetry)
        self._telemetry_timestamps[source] = telemetry_time

        bad_status = set(('offline', 'error'))
        if len(bad_status.intersection(device_status)) != 0:
            self._telemetry_pkg['state'] = 'ERR'
            self._telemetry_pkg['state_reason'] = '{} offline or errored'.format(source)
        else:
            state, reason = self._check_state()
            self._telemetry_pkg['state'] = state
            self._telemetry_pkg['state_reason'] = reason

        if source == 'flow':
            # report from flow_sensor
            pen_name = 'pen' + self.pen_number
            value = telemetry[pen_name]
            self._telemetry_pkg['flow'] = value
        elif source == 'pressure':
            self._telemetry_pkg['pressure'] = telemetry['pressure']
        elif source == 'peltier':
            new_tel = {}
            new_tel['heatsink'] = telemetry['heatsink' + self.meerstetter_channel]
            new_tel['peltier_current'] = telemetry['current' + self.meerstetter_channel]
            new_tel['peltier_status'] = telemetry['status' + self.meerstetter_channel]
            new_tel['setpoint'] = telemetry['target_temperature' + self.meerstetter_channel]
            new_tel['temperature'] = telemetry['temperature' + self.meerstetter_channel]
            self._telemetry_pkg.update(new_tel)

    def telemetry(self):
        # use stalest timestamp we have as the data timestamp
        oldest_ts = min(self._telemetry_timestamps,
                        key=self._telemetry_timestamps.get)
        tel = dict(timestamp=self._telemetry_timestamps[oldest_ts])
        tel.update(self._telemetry_pkg)
        return tel



from astropy import units as u

microstep = u.def_unit('microstep')


class DeviceProperty(object):
    """
    A descriptor for access to a property of a Device

    For example, one might wish to get or set the acceleration of a motor
    stage. There is a fair amount of boiler-plate code involved in getting
    or setting this property, but all that differs between them is the command
    code used and the units of the property.

    This is a perfect use of a Python descriptor
    """
    def __init__(self, code, units=u.dimensionless_unscaled,
                 relative_units=True, errcheck=True, readonly=False):
        """
        Parameters
        ----------
        code : string
            command code used to get or set property, e.g QS for microstep
        units : ```~astropy.units.Quantity```
            the units of the property, e.g ```~astropy.unit.A``` for current.
            see also `relative_units`.
        relative_units : bool
            units are in terms of Device base units.

            Some properties are always returned in terms of the Device' base units.
            For example,  the velocity is returned in 'units'/sec, where 'units' can
            be set on the controller Device. If `relative_units` is `True`, the property
            units are given by the controller base units multiplied by the `unit` of the
            property.
        errcheck : bool
            check for errors on get/set
        readonly : bool
            read only property
        """
        self.code = code
        self.units = units
        self.relative_units = relative_units
        self.errcheck = errcheck
        self.readonly = readonly

    def __get__(self, instance, cls):
        if instance is None:
            # called from class itself
            return self
        else:
            # called from class instance
            code = instance._format_get_command(self.code)
            val = float(instance._cmd(code, errcheck=self.errcheck))
            if self.relative_units:
                units = instance.units * self.units
            else:
                units = self.units
            return val * units

    def __set__(self, instance, value):
        if self.readonly:
            raise AttributeError("can't set attribute")

        # if we are passed a float or int, assume correct units and create quantity
        if not isinstance(value, u.Quantity):
            value = value * self.units * instance.units if self.relative_units else value * self.units

        # get value in correct units
        value = value.to_value(
            self.units * instance.units if self.relative_units else self.units
        )

        # set value
        code = instance._format_set_command(self.code)
        instance._cmd(code, params=(value,), errcheck=self.errcheck)


class IntegerDeviceProperty(DeviceProperty):
    """
    An Device property that must be integer.
    """
    def __init__(self, code, errcheck=True, readonly=False):
        super(IntegerDeviceProperty, self).__init__(code,
                                                    units=u.dimensionless_unscaled,
                                                    relative_units=False,
                                                    errcheck=errcheck, readonly=readonly)

    def __get__(self, instance, cls):
        value = super(IntegerDeviceProperty, self).__get__(instance, cls)
        return int(value)

    def __set__(self, instance, value):
        super(IntegerDeviceProperty, self).__set__(instance, int(value))


class BooleanDeviceProperty(DeviceProperty):
    """
    An Device property that must be boolean.
    """
    def __init__(self, code, errcheck=True, readonly=False):
        super(BooleanDeviceProperty, self).__init__(code,
                                                    units=u.dimensionless_unscaled,
                                                    relative_units=False,
                                                    errcheck=errcheck, readonly=readonly)

    def __get__(self, instance, cls):
        value = super(BooleanDeviceProperty, self).__get__(instance, cls).value
        return bool(value)

    def __set__(self, instance, value):
        super(BooleanDeviceProperty, self).__set__(instance, bool(value))



#!/usr/bin/env python
"""
Zaber Linear Stages, including T-LLS devices
"""
from __future__ import print_function, division, absolute_import
import struct
import six
import socket
import threading
import time

from .properties import DeviceProperty, BooleanDeviceProperty, microstep

# rename socket.error
if six.PY2:
    ConnectionRefusedError = socket.error


class ZaberError(Exception):
    pass


class ZaberTimeoutError(ZaberError):
    pass


# how long to wait before giving up on a response
DEFAULT_TIMEOUT = 5

# number of bytes to transmit & recieve
PACKET_SIZE = 6

# command numbers
RESET = 0
HOME = 1
MOVE_ABSOLUTE = 20
MOVE_RELATIVE = 21
STOP = 23
RESTORE = 36
SET_MODE = 40
MAX_POSITION = 44
RETURN_SETTING = 53
RETURN_STATUS = 54
POSITION = 60
NULL = 0

# error return from slide
ERROR = 255

# unit number, may depend upon how device is connected to the port
UNIT = 1

PERIPHERAL_ID = 0
POTENTIOM_OFF = 8
POTENTIOM_ON = 0

TRUE = 1
FALSE = 0


class ZaberLS(object):
    position = DeviceProperty(POSITION, readonly=True)
    moving = BooleanDeviceProperty(RETURN_STATUS, readonly=True)

    def __init__(self, host="192.168.1.3", port=10001, unit_equivalencies=None):
        """
        Creates a Zaber device. Arguments::

         host: IP address of terminal server
         port: port device representing the slide

        """
        self.port = port
        self.host = host
        # thread lock for threadsafe operations
        self._lock = threading.Lock()
        self.unit = UNIT
        self.unit_equivalencies = unit_equivalencies
        self.units = microstep
        self.sock = None

    def __del__(self):
        # must close connection on object deletion
        self.close()

    def connect(self):
        # connect to socket on terminal server
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.host, self.port))
        self.sock = s
        self.sock.settimeout(DEFAULT_TIMEOUT)

    def close(self):
        if self.sock is not None:
            self.sock.close()
        self.sock = None

    def _format_get_command(self, code):
        return code

    def _format_set_command(self, code):
        return code

    def _cmd(self, cmd, params=NULL, errcheck=True):
        # set command data
        to_send = self._encodeCommandData(params)
        # add bytes to define instruction at start of array
        to_send.insert(0, cmd)
        # add unit byte
        to_send.insert(0, self.unit)

        received = self._sendRecv(to_send)
        value = self._decodeCommandData(received)

        if received[1] == ERROR and errcheck:
            # spontaneous error reply
            raise ZaberError("Error occured with code: {}".format(value))
        return value

    def _sendRecv(self, byteArr):
        # only one thread at a time talks to the slide
        with self._lock:
            try:
                self.sock.send(byteArr)
            except socket.timeout:
                raise ZaberTimeoutError(
                    "timed out sending data to slide. this is normal on long moves"
                )
            except Exception as e:
                raise ZaberError("failed to send bytes to slide" + str(e))

            # now try and receive
            start_time = time.time()
            msg = b""
            while len(msg) < 6 and time.time() - start_time < DEFAULT_TIMEOUT:
                try:
                    chunk = self.sock.recv(6 - len(msg))
                    msg += chunk
                except socket.timeout:
                    raise ZaberTimeoutError(
                        "timed out getting response from slide. normal on long moves."
                    )
                except Exception as e:
                    raise ZaberError("failed to receive bytes to slide" + str(e))
            if time.time() - start_time >= DEFAULT_TIMEOUT:
                raise ZaberTimeoutError("timed out gettin response from slide")
        return bytearray(msg)

    def _decodeCommandData(self, byteArr):
        return struct.unpack("<L", byteArr[2:])[0]

    def _encodeCommandData(self, int):
        return bytearray(struct.pack("<L", int))

    def _encodeByteArr(self, intArr):
        if len(intArr) != 6:
            raise ZaberError("must send 6 bytes to slide: cannot send " + repr(intArr))
        return bytearray(intArr)

    @property
    def homed(self):
        """
        returns true if the slide has been homed and has a calibrated
        position
        """
        byteArr = self._encodeByteArr(
            [UNIT, RETURN_SETTING, SET_MODE, NULL, NULL, NULL]
        )
        byteArr = self._sendRecv(byteArr)
        if byteArr[1] == ERROR:
            raise ZaberError("Error trying to get the setting byte")

        # if 7th bit is set, we have been homed
        if byteArr[2] & 128:
            return True
        else:
            return False

    def move(self, position):
        """
        move to a defined position
        """
        if hasattr(position, "unit"):
            nstep = int(
                position.to(microstep, equivalencies=self.unit_equivalencies).value
            )
        else:
            # not a unit, assume microsteps
            nstep = int(position)

        # valid position?
        if nstep < 0 or nstep > self.max_position:
            raise ZaberError(
                "Cannot move to position {} as it is outside the range of the slide ({}-{})".format(
                    nstep, 0, self.max_position
                )
            )

        self._cmd(MOVE_ABSOLUTE, nstep, errcheck=True)

    def home(self):
        """
        move the slide to the home position. This is needed after a power on
        to calibrate the slide
        """
        self._cmd(HOME)

    def reset(self):
        """
        carry out the reset command, equivalent to turning the slide off and
        on again. The position of the slide will be lost and a home will be
        needed
        """
        self._cmd(RESET)

    def restore(self):
        """
        carry out the restore command. restores the device to factory settings
        very useful if the device does not appear to function correctly
        """
        self._cmd(RESTORE, params=PERIPHERAL_ID)

    def disable(self):
        """
        carry out the disable command. disables the potentiometer preventing
        manual adjustment of the device
        """
        self._cmd(SET_MODE, POTENTIOM_OFF)

    def enable(self):
        """
        carry out the enable command. enables the potentiometer allowing
        manual adjustment of the device
        """
        self._cmd(SET_MODE, POTENTIOM_ON)

    def stop(self):
        """stop the slide"""
        self._cmd(STOP)


class ZaberLSEmulator(ZaberLS):
    def __init__(self, host="192.168.1.3", port=10001, unit_equivalencies=None):
        super(ZaberLSEmulator, self).__init__(host, port, unit_equivalencies)
        self.target_position = None
        self.current_position = None
        self._homed = False
        self._moving = False
        self._running_thread = None
        self.abort_thread = False

    @property
    def homed(self):
        return self._homed

    @homed.setter
    def homed(self, value):
        self._homed = value

    @property
    def moving(self):
        return self._moving

    @moving.setter
    def moving(self, value):
        self._moving = value

    def connect(self):
        pass

    def close(self):
        pass

    def stop(self):
        self.abort_thread = True
        self._moving = False

    def _cmd(self, cmd, params=NULL, errcheck=True):
        response_data = params

        if cmd == HOME:
            self.target_position = 0 * microstep
            if self.current_position is None:
                self.current_position = 19000 * microstep
            self.homed = True
            self.moving = True
            t = threading.Thread(target=self._move_thread)
            t.start()
            self._running_thread = t

        elif cmd == MOVE_ABSOLUTE:
            self.target_position = params * microstep
            self.moving = True

            t = threading.Thread(target=self._move_thread)
            t.start()
            self._running_thread = t

        elif cmd == POSITION:
            response_data = int(self.current_position.value)

        # set command data
        to_send = self._encodeCommandData(response_data)
        # add bytes to define instruction at start of array
        to_send.insert(0, cmd)
        # add unit byte
        to_send.insert(0, self.unit)
        return self._decodeCommandData(to_send)

    def _move_thread(self):
        self.abort_thread = False
        speed = 9000  # steps/s
        step_time = 0.1  # s
        nsteps = int(
            abs((self.target_position - self.current_position).value)
            / speed
            / step_time
        )
        if nsteps == 0:
            step = 0
        else:
            step = (self.target_position - self.current_position) / nsteps
        for _ in range(nsteps):
            self.current_position += step
            time.sleep(step_time)
            if self.abort_thread:
                break
        self.moving = False



import astropy.units as u
from astropy.time import Time
from twisted.logger import Logger

from ..devices.pdr900 import PDR900, PDR900Emulator
from ..machines import create_machine

torr = u.def_unit('torr', (101325/760)*u.Pa)
UNITS = u.mbar


class VacGauge(object):
    """
    Model for PDR900 vacuum gauge.

    This is a model for a single vacuum gauge.
    """
    log = Logger()

    settable_attributes = ()
    extra_rpc_calls = ()

    def __init__(self, host, port, emulate=True):
        if emulate:
            self.dev = PDR900Emulator(1.0e-5, 0.0)
        else:
            self.dev = PDR900(host, port)

        # state machine to keep track of connections
        self._controller = create_machine('connections',
                                          initial_context={'device': self, 'name': 'vac gauge'})
        self.machines = {'connection': self._controller}

        ctx = {
            'value': 0.0,
            'read_value': lambda *args: self.dev.pressure,
            'name': 'vac_gauge'
        }
        sensor = create_machine('sensor', initial_context=ctx)
        self.gauge = sensor
        self.machines['gauge'] = sensor
        # sensors will recieve connect/disconnect events from controller
        self._controller.bind(sensor)
        sensor.bind(self._controller)

    def try_connect(self):
        try:
            self.dev.connect()
        except Exception:
            return False
        return True

    def disconnect(self):
        self.dev.close()

    def telemetry(self):
        """
        Called periodically to provide a telemetry package for the device
        """
        ts = Time(self.gauge.context['last_read'], format='unix')
        ts.format = 'iso'
        tel = dict(
            timestamp=ts,
            state={key: self.machines[key].configuration for key in self.machines}
        )
        tel['pressure'] = (self.gauge.context['value'] * UNITS).to(u.mbar)
        return tel



import time
import numpy as np


class SensorEmulator(object):
    """
    Simple class to emulate a device sensor

    Noise is added as a Gaussian Random Walk
    """
    def __init__(self, initial_value, response_speed, noise):
        self._current = initial_value
        self._target = initial_value
        self.speed = response_speed
        self.noise = noise
        self._time_check = time.time()

    @property
    def current_value(self):
        deltaT = time.time() - self._time_check
        value = abs((deltaT * self.speed))
        if value > abs(self._target - self._current):
            value = abs(self._target - self._current)

        if self._current > self._target:
            value *= -1

        self._current += value + np.random.normal(0, self.noise)
        self._time_check = time.time()
        return self._current

    @property
    def target_value(self):
        return self._target

    @target_value.setter
    def target_value(self, value):
        self._time_check = time.time()
        self._target = value



# focal plane slide
import astropy.units as u
from astropy.time import Time
from twisted.logger import Logger

from ..devices.properties import microstep
from ..devices.zaber import ZaberLS, ZaberLSEmulator, ZaberTimeoutError
from ..machines import create_machine

# the next define ranges for the movement in terms of
# microsteps, millimetres and pixels
MIN_MS = 0
MAX_MS = 672000
MIN_PX = 1453.1212
MAX_PX = -125.0606
MM_PER_MS = 0.00015625
# END VIK CHANGE ME!
MIN_MM = MM_PER_MS * MIN_MS
MAX_MM = MM_PER_MS * MAX_MS

# Standard pixel positions for unblocking and blocking the CCD
UNBLOCK_POS = 1100.0
BLOCK_POS = -100.0


def mm_to_ms(value):
    return MIN_MS + int((MAX_MS - MIN_MS) * (value - MIN_MM) / (MAX_MM - MIN_MM) + 0.5)


def ms_to_mm(value):
    return MIN_MM + (MAX_MM - MIN_MM) * (value - MIN_MS) / (MAX_MS - MIN_MS)


def px_to_ms(value):
    return MIN_MS + int((MAX_MS - MIN_MS) * (value - MIN_PX) / (MAX_PX - MIN_PX) + 0.5)


def ms_to_px(value):
    return MIN_PX + (MAX_PX - MIN_PX) * (value - MIN_MS) / (MAX_MS - MIN_MS)


fps_scale = [
    # from, to, forward, backward
    (u.pix, microstep, px_to_ms, ms_to_px),
    (u.mm, microstep, mm_to_ms, ms_to_mm),
]


class LinearStage(object):
    """
    Zaber linear stage.

    Uses a Motor state machine to track state, and a Zaber linear stage device to send
    messages.

    Periodically updates WAMP server with updates of position and state.
    """

    log = Logger()

    # can be set by publishing to device level topic of same name
    settable_attributes = ("target_position",)
    extra_rpc_calls = ()

    def __init__(self, ip_address, port, equivalencies, emulate=True):
        # unit handling
        self.equivalencies = equivalencies
        # units used for public facing reports
        self.natural_units = microstep

        if emulate:
            self.dev = ZaberLSEmulator(
                ip_address, port, unit_equivalencies=equivalencies
            )
        else:
            self.dev = ZaberLS(ip_address, port, unit_equivalencies=equivalencies)

        # override the maximum position to avoid querying it, which is unreliable
        self.dev.max_position = 672255

        # state machines to keep track of state
        self._controller = create_machine(
            "connections", initial_context={"device": self, "name": "slide"}
        )
        self._axis = create_machine(
            "motors",
            initial_context={
                "device": self,
                "current": 0 * self.natural_units,
                "target": 0 * self.natural_units,
                "poll_time": 1,
            },
        )

        self.machines = {"connection": self._controller, "stage": self._axis}
        # axis will recieve events from controller
        self._controller.bind(self._axis)

    @property
    def current_position(self):
        with u.set_enabled_equivalencies(self.equivalencies):
            return self._axis.context["current"].to(self.natural_units)

    @property
    def target_position(self):
        with u.set_enabled_equivalencies(self.equivalencies):
            return self._axis.context["target"].to(self.natural_units)

    @target_position.setter
    def target_position(self, value):
        if not isinstance(value, u.Quantity):
            value = value * self.natural_units
        self._axis.queue("positionSet", position=value)

    def telemetry(self):
        """
        Called periodically to provide a telemetry package for the device
        """
        ts = Time.now()
        return dict(
            timestamp=ts,
            position=dict(current=self.current_position, target=self.target_position),
            state={key: self.machines[key].configuration for key in self.machines},
        )

    # methods and properties required for motor state machine
    # Keep these on slide object rather than moving to Zaber since
    # querying target position of Zaber stage is not implemented
    def home(self):
        try:
            self.dev.home()
        except ZaberTimeoutError as err:
            # assume if reply times out it's fine
            self.log.error("timeout " + str(err))
            return True
        except Exception as err:
            self.log.error(str(err))
            return False
        return True

    def homed(self):
        is_homed = False
        try:
            is_homed = self.dev.homed
        except Exception as err:
            self.log.error(str(err))
        return is_homed

    def on_target(self):
        try:
            current = self.current_position
            target = self.target_position
            if not isinstance(current, u.Quantity):
                current = current * self.natural_units
            if not isinstance(target, u.Quantity):
                target = target * self.natural_units
            current = self.current_position.to(
                microstep, equivalencies=self.equivalencies
            )
            target = self.target_position.to(
                microstep, equivalencies=self.equivalencies
            )
            delta = abs((current - target).value)
            return int(delta) < 5
        except Exception as err:
            self.log.error(str(err))
            return False

    def move(self, value):
        if not isinstance(value, u.Quantity):
            # no units, assume natural units
            value = value * self.natural_units
        try:
            self.dev.move(value)
        except ZaberTimeoutError as err:
            # assume if reply times out it's fine
            self.log.error("timeout " + str(err))
            return True
        except Exception as err:
            self.log.error(str(err))
            return False
        return True

    def stop(self):
        try:
            self.dev.stop()
        except ZaberTimeoutError as err:
            # assume if reply times out it's fine
            self.log.error("timeout " + str(err))
            return True
        except Exception as err:
            self.log.error(str(err))
            return False
        return True

    def try_connect(self):
        """
        Always needs to be defined.  Returns True on successful connection attempt.
        """
        # attempt connection, return true to allow device state to change
        try:
            self.dev.connect()
        except Exception:
            return False
        return True

    def disconnect(self):
        self.dev.close()

    def poll(self):
        """
        State machines mean that this is automatically polled.

        Update current position and return True/False for moving state
        """
        try:
            current_position = self.dev.position.to(
                self.natural_units, equivalencies=self.equivalencies
            )
            moving = self.dev.moving
        except Exception as err:
            self.log.error(str(err))
            current_position = 0 * self.natural_units
            moving = True
        return current_position, self.dev.moving


class FocalPlaneSlide(LinearStage):
    """
    Focal plane slide.

    Uses the unit equivalencies defined above and adds "block" etc as RPC calls

    All positions in pixels
    """

    # all state change triggers are automatically addded as RPC calls
    # add any extra methods you want exposed here
    extra_rpc_calls = ("block", "unblock", "enable", "disable", "reset")

    def __init__(self, ip_address, port, emulate=True):
        super().__init__(ip_address, port, fps_scale, emulate)
        self.natural_units = u.pix

    def block(self):
        self.target_position = BLOCK_POS
        self._axis.queue("move")

    def unblock(self):
        self.target_position = UNBLOCK_POS
        self._axis.queue("move")

    def enable(self):
        self.dev.enable()

    def disable(self):
        self.dev.disable()

    def reset(self):
        self.dev.reset()
        self._controller.queue("disconnect")



from functools import partial

from astropy import units as u
from astropy.time import Time
from twisted.logger import Logger

from ..devices.meerstetter import (MeerstetterTEC1090, STATUS_CODES,
                                   MeerstetterTEC1090Emulator)
from ..machines import create_machine

TEMP_UNITS = u.deg_C
CURRENT_UNITS = u.A


# define some helper functions to create sensor statemachines
def create_sensor(connection_handler, address, name, read_method, initial_value):
    sensor = create_machine('sensor',
                            initial_context={
                                'value': initial_value,
                                'read_value': read_method,
                                'name': name
                            })
    connection_handler.bind(sensor)
    sensor.bind(connection_handler)
    return sensor


def create_controller(connection_handler, address, name, read_method,
                      write_method, setpoint_read_method):
    sensor = create_machine('controller',
                            initial_context={
                                'write': write_method,
                                'read': read_method,
                                'name': name,
                                'read_setpoint': setpoint_read_method
                            })
    connection_handler.bind(sensor)
    sensor.bind(connection_handler)
    return sensor


class Meerstetter(object):
    """
    Meerstetter peltier controller.

    This is a model for a single peltier controller, that has three channels.
    """
    log = Logger()

    # can be set by publishing to device level topic of same name
    settable_attributes = ('target_temperature1', 'target_temperature2',
                           'target_temperature3')

    # all state change triggers are automatically addded as RPC calls
    # add any extra methods you want exposed here
    extra_rpc_calls = ('reset',)

    def __init__(self, ip_addr, port, emulate=True):

        if emulate:
            dev = MeerstetterTEC1090Emulator(ip_addr, port)
            self.dev = dev
        else:
            dev = MeerstetterTEC1090(ip_addr, port)
            self.dev = dev

        self._controller = create_machine('connections',
                                          initial_context={'device': self,
                                                           'name': 'meerstetter'})
        self.machines = {'connection': self._controller}

        for addr in range(1, 4):
            name = 'status{}'.format(addr)
            self.machines[name] = create_sensor(
                self._controller, addr, name, partial(self.dev.get_status, addr), 0.0)

            name = 'heatsink{}'.format(addr)
            self.machines[name] = create_sensor(
                self._controller, addr, name,
                partial(self.dev.get_heatsink_temp, addr), 0.0)

            name = 'current{}'.format(addr)
            self.machines[name] = create_sensor(
                self._controller, addr, name, partial(self.dev.get_current, addr), 0.0)

            name = 'temperature{}'.format(addr)
            self.machines[name] = create_controller(
                self._controller, addr, name,
                partial(self.dev.get_ccd_temp, addr),
                partial(self.dev.set_ccd_temp, addr),
                partial(self.dev.get_setpoint, addr))

    @property
    def target_temperature1(self):
        return self.machines['temperature1'].context['target']

    @target_temperature1.setter
    def target_temperature1(self, value):
        """
        Set target. Assume raw floats are already in degrees
        """
        if isinstance(value, u.Quantity):
            value = value.to_value(TEMP_UNITS, equivalencies=u.temperature())
        self.machines['temperature1'].queue('targetSet', target=value)

    @property
    def target_temperature2(self):
        return self.machines['temperature2'].context['target']

    @target_temperature2.setter
    def target_temperature2(self, value):
        if isinstance(value, u.Quantity):
            value = value.to_value(TEMP_UNITS, equivalencies=u.temperature())
        self.machines['temperature2'].queue('targetSet', target=value)

    @property
    def target_temperature3(self):
        return self.machines['temperature3'].context['target']

    @target_temperature3.setter
    def target_temperature3(self, value):
        if isinstance(value, u.Quantity):
            value = value.to_value(TEMP_UNITS, equivalencies=u.temperature())
        self.machines['temperature3'].queue('targetSet', target=value)

    def reset(self, address):
        self.dev.reset_tec(address)

    def _oldest_timestamp(self):
        """
        search all the sensors and return the oldest (stalest) timestamp
        """
        stamps = []
        for key in self.machines:
            if 'last_read' in self.machines[key].context:
                stamps.append(self.machines[key].context['last_read'])
        return min(stamps)

    def telemetry(self):
        ts = Time(self._oldest_timestamp(), format='unix')
        ts.format = 'iso'
        tel = dict(
            timestamp=ts,
            state={key: self.machines[key].configuration for key in self.machines}
        )
        for key in self.machines:
            if key.startswith('status'):
                tel[key] = STATUS_CODES[self.machines[key].context['value']]
            elif key != 'connection':
                unit = CURRENT_UNITS if key.startswith('current') else TEMP_UNITS
                tel[key] = self.machines[key].context['value'] * unit
        for addr in range(1, 4):
            key = 'target_temperature{}'.format(addr)
            tel[key] = getattr(self, key) * TEMP_UNITS
        return tel

    # METHODS NEEDED FOR STATE MACHINES
    def try_connect(self):
        """
        Always needs to be defined.  Returns True on successful connection attempt.
        """
        # attempt connection, return true to allow device state to change
        try:
            connected = self.dev.connect()
        except Exception:
            return False
        return connected

    def disconnect(self):
        self.dev.disconnect()



from functools import partial
import astropy.units as u
from astropy.time import Time
from twisted.logger import Logger

from ..devices.honeywell import Minitrend, MinitrendEmulator
from ..machines import create_machine


UNITS = u.liter/u.minute


class FlowSensor(object):
    """
    Honeywell Minitrend Flow Sensor

    All flow rates in liters/minute
    """
    log = Logger()

    settable_attributes = ()
    extra_rpc_calls = ()

    def __init__(self, ip_address, port, npens=6, emulate=True):
        if emulate:
            self.dev = MinitrendEmulator(1.0, 0.05)
        else:
            self.dev = Minitrend(ip_address, port)

        # state machine to keep track of state
        self._controller = create_machine('connections',
                                          initial_context={'device': self, 'name': 'honeywell'})

        self.machines = {'connection': self._controller}
        self.npens = npens
        # map pens to ones used in hipercam
        self.pen_mapping = {0: 0, 1: 1, 2: 2, 3: 3, 4: 8, 5: 9}
        self.sensors = []

        def make_alarm_read(pen):
            def inner():
                alarm = any(self.dev.get_alarm(pen))
                return alarm
            return inner

        for i in range(self.npens):
            pen_num = self.pen_mapping[i]
            ctx = {
                'value': 0.0,
                'read_value': partial(self.dev.get_pen, pen_num),
                'read_alarm': make_alarm_read(pen_num),
                'name': 'flow{}'.format(pen_num)
            }
            sensor = create_machine('sensor', initial_context=ctx)
            self.machines['pen' + str(i+1)] = sensor
            # sensors will recieve connect/disconnect events from controller
            # and vice versa
            self._controller.bind(sensor)
            sensor.bind(self._controller)

    def try_connect(self):
        try:
            self.dev.connect()
        except Exception:
            return False
        return True

    def disconnect(self):
        self.dev.close()

    def _oldest_timestamp(self):
        """
        search all the sensors and return the oldest (stalest) timestamp
        """
        stamps = []
        for key in self.machines:
            if 'last_read' in self.machines[key].context:
                stamps.append(self.machines[key].context['last_read'])
        return min(stamps)

    def telemetry(self):
        """
        Called periodically to provide a telemetry package for the device
        """
        ts = Time(self._oldest_timestamp(), format='unix')
        ts.format = 'iso'
        tel = dict(timestamp=ts)
        tel['state'] = {key: self.machines[key].configuration for key in self.machines}
        for i in range(self.npens):
            name = 'pen' + str(i+1)
            sensor = self.machines[name]
            tel[name] = sensor.context['value'] * UNITS
            tel[name.replace('pen', 'alarm')] = sensor.context['alarm']
        return tel



import os
import time
import glob

from watchdog.utils.dirsnapshot import DirectorySnapshot, DirectorySnapshotDiff
from watchdog.events import PatternMatchingEventHandler
from watchdog.events import (
    FileCreatedEvent,
    FileModifiedEvent
)


class FITSWriteHandler(PatternMatchingEventHandler):
    """
    Class to handle FITS write events.abs

    Upon write events to the current run, we send offset to
    the telescope, schedule callbacks to see if the
    offset is complete, and then send a trigger to the
    sequencer to continue with the observation.

    Arguments
    ----------
    path : str
        location of fits files to monitor
    """
    patterns = ['*.fits']

    def __init__(self, path, on_modified_callback, *args, **kwargs):
        super(FITSWriteHandler, self).__init__()
        self.path = path
        self.on_modified_callback = on_modified_callback
        try:
            pattern = os.path.abspath(os.path.join(self.path, self.patterns[0]))
            self.existing_runs = glob.glob(pattern)
        except Exception:
            self.existing_runs = []

        # used to avoid responding to very recent
        # events which may be duplicates.
        self.debounce_time = 0.1
        self.last_event = 0

    def check_debounce(self):
        """
        Avoid handling very closely occurring filesystem events.abs

        Sometimes there are two or more rapid filesystem events when
        a frame is written. This routine checks if another event
        as happened recently, and returns False if so.
        """
        first_event = False
        if time.time() - self.last_event > self.debounce_time:
            self.last_event = time.time()
            first_event = True
        else:
            self.last_event = time.time()
        return first_event

    def on_modified(self, event):
        if self.check_debounce():
            if os.path.abspath(event.src_path) == self.existing_runs[-1]:
                # latest file has just been written to
                self.on_modified_callback()

    def on_created(self, event):
        self.check_debounce()
        # if we don't know about this, it's a new run.
        if os.path.abspath(event.src_path) not in self.existing_runs:
            self.existing_runs.append(event.src_path)
            # log to window
            run_file = os.path.split(event.src_path)[-1]
            run_name = os.path.splitext(run_file)[0]
            print('HiPERCAM is observing new run: {}'.format(run_name))


class FITSWatcher(object):

    def __init__(self, path, on_modified_callback):
        self.path = path
        self.handler = FITSWriteHandler(self.path, on_modified_callback)
        self._snapshot = self.take_snapshot()

    def take_snapshot(self):
        return DirectorySnapshot(self.path, recursive=False)

    def poll(self):
        new_snapshot = self.take_snapshot()
        events = DirectorySnapshotDiff(self._snapshot, new_snapshot)
        self._snapshot = new_snapshot

        # process file creation and modified events only
        for src_path in events.files_created:
            self.queue_event(FileCreatedEvent(src_path))
        for src_path in events.files_modified:
            self.queue_event(FileModifiedEvent(src_path))

    def queue_event(self, event):
        self.handler.dispatch(event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import glob
import os
import sys

with open("README.rst", encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst", encoding="utf-8") as history_file:
    history = history_file.read()

requirements = [
    "astropy",
    "requests",
    "sismic",
    "autobahn",
    "pyserial",
    "twisted",
    "influxdb",
    "txaio",
    "watchdog",
    "six",
]

# for Python 3.6 we need pymodbus3
modern_requirements = ["pymodbus"]
older_requirements = ["pymodbus3"]

test_requirements = [
    # TODO: put package test requirements here
]

# Treat everything in scripts except README.rst as a script to be installed
scripts = [
    fname
    for fname in glob.glob(os.path.join("scripts", "*"))
    if os.path.basename(fname) != "README.rst"
]

setup(
    name="hcam_devices",
    version="1.2.1",
    description="Device Communication via WAMP for HiPerCAM",
    long_description=readme + "\n\n" + history,
    author="Stuart Littlefair",
    author_email="s.littlefair@shef.ac.uk",
    url="https://github.com/HiPERCAM/hcam_devices",
    download_url="https://github.com/HiPERCAM/hcam_devices/archive/v1.2.1.tar.gz",
    packages=[
        "hcam_devices",
        "hcam_devices.gtc",
        "hcam_devices.devices",
        "hcam_devices.components",
        "hcam_devices.machines",
        "hcam_devices.models",
        "hcam_devices.testing",
        "hcam_devices.wamp",
        "hcam_devices.wamp.utils",
        "hcam_devices.utils",
    ],
    package_dir={"hcam_devices": "hcam_devices"},
    include_package_data=True,
    scripts=scripts,
    install_requires=requirements,
    extras_require={
        ':python_version == "3.6"': older_requirements,
        ':python_version >= "3.7"': modern_requirements,
    },
    license="MIT license",
    zip_safe=False,
    keywords="hcam_devices",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
    test_suite="tests",
    tests_require=test_requirements,
)
